import numpy as np

# from .pushers import factory_pushers


def verify_value(
    x,
    val_shape,
    mom_shape=None,
    axis=None,
    nrec=None,
    expand=False,
    broadcast=False,
    flatten=True,
    verify=False,
    dtype=None,
    order=None,
    shape=None,
    shape_flat=None,
    first=True,
):
    """
    given an array `x` verify that it conforms to a target shape

    Parameters
    ----------
    x: array-like
        array to consider
    val_shape : tuple
        shape of value part of target
    mom_shape : tuple, optional
        shape of moment part of target.  If None, then target shape
        excludes moments
    verify: callable, optional
        if present, `x = verify(x)`.  For example, can pass
        verify=numpy.asarray


    axis: int, optional
        if specified, then this is the axis that will be reduced over.
        This axis will be moved to the first axis of the output
    nrec: int, optional
        number of records.  That is, x.shape[axis]
        Pass this to perform a check on sizes.
    broadcast: bool, default=False
        if True, then broadcast to target shape

    """

    if verify:
        x = np.asarray(x, dtype=dtype, order=order)

    ndim = len(val_shape)
    axis = np.core.numeric.normalize_axis_index(axis, ndim + 1)

    if shape is None or shape_flat is None:
        if ndim == 0:
            shape_flat = ()
        else:
            shape_flat = (np.prod(shape),)

        if axis is not None:
            shape = shape[:axis] + (nrec,) + shape[axis:]
            shape_flat = (nrec,) + shape_flat

        if mom_shape is not None:
            shape = shape + mom_shape
            shape_flat = shape_flat + mom_shape

    if expand:
        if x.ndim == 1 and x.ndim != len(shape) and len(x) == shape[axis]:

            reshape = [1] * len(shape)
            reshape[axis] = -1
            x = x.reshape(*reshape)

    if broadcast and x.shape != shape:
        x = np.broadcast_to(x, shape)
    else:
        assert x.shape == shape

    if first or flatten and axis != 0:
        x = np.moveaxis(x, axis, 0)

    if flatten:
        x = x.reshape(shape_flat)

    if x.ndim == 0:
        x = x[()]

    return x, shape, shape_flat


# def push_val(x,
#              mom,
#              w=None,
#              verify=True,
#              broadcast=False,
#              data=None,
#              data_flat=None,
#              val_shape=None,
# ):


#     if isinstance(mom, int):
#         mom = (mom,)
#     mom_ndim = len(mom)
#     assert mom_ndim in (1, 2)
#     mom_shape = tuple(x+1 for x in mom)


#     if mom_ndim == 1:
#         x = (x,)

#     if verify:
#         x = (np.asarray(_) for _ in x)

#     if val_shape is None:
#         val_shape = x[0].shape

#     if val_shape == ():
#         val_shape_flat = ()
#     else:
#         val_shape_flat = (np.prod(val_shape), )

#     shape = val_shape + mom_shape
#     shape_flat = val_shape_flat + mom_shape

#     # verify x[0]
#     x = verify_value(x, val_shape=val_shape, mom_shape=)



"""
routines to convert central (co)moments to raw (co)moments
"""
from __future__ import absolute_import

import numpy as np

from .options import OPTIONS
from .utils import factory_binomial, myjit

_bfac = factory_binomial(OPTIONS["nmax"])


@myjit
def _central_to_raw_moments(central, raw):
    nv = central.shape[0]
    order = central.shape[1] - 1

    for v in range(nv):
        c = central[v]
        r = raw[v]

        ave = c[1]

        r[0] = c[0]
        r[1] = ave

        for n in range(2, order + 1):
            tmp = 0.0
            ave_i = 1.0
            for i in range(0, n - 1):
                tmp += c[n - i] * ave_i * _bfac[n, i]
                ave_i *= ave

            # last two
            # <dx> = 0 so skip i = n-1
            # i = n
            tmp += ave_i * ave
            r[n] = tmp


@myjit
def _raw_to_central_moments(raw, central):
    nv = central.shape[0]
    order = central.shape[1] - 1

    for v in range(nv):
        c = central[v]
        r = raw[v]

        ave = r[1]

        c[0] = r[0]
        c[1] = ave

        for n in range(2, order + 1):
            tmp = 0.0
            ave_i = 1.0
            for i in range(0, n - 1):
                tmp += r[n - i] * ave_i * _bfac[n, i]
                ave_i *= -ave

            # last two
            # right now, ave_i = (-ave)**(n-1)
            # i = n-1
            # ave * ave_i * n
            # i = n
            # 1 * (-ave) * ave_i
            tmp += ave * ave_i * (n - 1)
            c[n] = tmp


# comoments
@myjit
def _central_to_raw_comoments(central, raw):
    nv = central.shape[0]
    order0 = central.shape[1] - 1
    order1 = central.shape[2] - 1

    for v in range(nv):
        c = central[v]
        r = raw[v]

        ave0 = c[1, 0]
        ave1 = c[0, 1]

        for n in range(0, order0 + 1):
            for m in range(0, order1 + 1):
                nm = n + m
                if nm <= 1:
                    r[n, m] = c[n, m]
                else:
                    tmp = 0.0
                    ave_i = 1.0
                    for i in range(n + 1):
                        ave_j = 1.0
                        for j in range(m + 1):
                            nm_ij = nm - (i + j)
                            if nm_ij == 0:
                                # both zero order
                                tmp += ave_i * ave_j
                            elif nm_ij == 1:
                                # <dx**0 * dy**1> = 0
                                pass
                            else:
                                tmp += (
                                    c[n - i, m - j]
                                    * ave_i
                                    * ave_j
                                    * _bfac[n, i]
                                    * _bfac[m, j]
                                )
                            ave_j *= ave1
                        ave_i *= ave0
                    r[n, m] = tmp


@myjit
def _raw_to_central_comoments(raw, central):
    nv = central.shape[0]
    order0 = central.shape[1] - 1
    order1 = central.shape[2] - 1

    for v in range(nv):
        c = central[v]
        r = raw[v]

        ave0 = r[1, 0]
        ave1 = r[0, 1]

        for n in range(0, order0 + 1):
            for m in range(0, order1 + 1):
                nm = n + m
                if nm <= 1:
                    c[n, m] = r[n, m]
                else:
                    tmp = 0.0
                    ave_i = 1.0
                    for i in range(n + 1):
                        ave_j = 1.0
                        for j in range(m + 1):
                            nm_ij = nm - (i + j)
                            if nm_ij == 0:
                                # both zero order
                                tmp += ave_i * ave_j
                            else:
                                tmp += (
                                    r[n - i, m - j]
                                    * ave_i
                                    * ave_j
                                    * _bfac[n, i]
                                    * _bfac[m, j]
                                )
                            ave_j *= -ave1
                        ave_i *= -ave0
                    c[n, m] = tmp


def _convert_moments(data, axis, target_axis, func, dtype=None, order=None, out=None):
    if isinstance(axis, int):
        axis = (axis,)
    if isinstance(target_axis, int):
        target_axis = (target_axis,)

    axis = tuple(axis)
    target_axis = tuple(target_axis)

    assert len(axis) == len(target_axis)

    data = np.asarray(data, dtype=dtype, order=order)
    if out is None:
        out = np.zeros_like(data)
    else:
        assert out.shape == data.shape
        out[...] = 0.0

    if axis != target_axis:
        data_r = np.moveaxis(data, axis, target_axis)
        out_r = np.moveaxis(out, axis, target_axis)
    else:
        data_r = data
        out_r = out

    shape = data_r.shape[: -len(axis)]
    if shape == ():
        reshape = (1,) + data_r.shape[-len(axis) :]
    else:
        reshape = (np.prod(shape),) + data_r.shape[-len(axis) :]

    data_r = data_r.reshape(reshape)
    out_r = out_r.reshape(reshape)

    func(data_r, out_r)
    return out


def to_raw_moments(x, axis=-1, dtype=None, order=None, out=None):
    """
    convert central moments to raw moments
    """
    if axis is None:
        axis = -1

    return _convert_moments(
        data=x,
        axis=axis,
        target_axis=-1,
        func=_central_to_raw_moments,
        dtype=dtype,
        order=order,
        out=out,
    )


def to_raw_comoments(x, axis=(-2, -1), dtype=None, order=None, out=None):
    """
    convert central moments to raw moments
    """

    if axis is None:
        axis = (-2, -1)

    return _convert_moments(
        data=x,
        axis=axis,
        target_axis=(-2, -1),
        func=_central_to_raw_comoments,
        dtype=dtype,
        order=order,
        out=out,
    )


def to_central_moments(x, axis=-1, dtype=None, order=None, out=None):
    """
    convert central moments to raw moments
    """

    if axis is None:
        axis = -1

    return _convert_moments(
        data=x,
        axis=axis,
        target_axis=-1,
        func=_raw_to_central_moments,
        dtype=dtype,
        order=order,
        out=out,
    )


def to_central_comoments(x, axis=(-2, -1), dtype=None, order=None, out=None):
    """
    convert raw comoments to central comoments
    """

    if axis is None:
        axis = (-2, -1)

    return _convert_moments(
        data=x,
        axis=axis,
        target_axis=(-2, -1),
        func=_raw_to_central_comoments,
        dtype=dtype,
        order=order,
        out=out,
    )



"""
This sets up optional values
"""

NMAX = "nmax"
FASTMATH = "fastmath"
PARALLEL = "parallel"
CACHE = "cache"


OPTIONS = {NMAX: 20, PARALLEL: True, CACHE: True, FASTMATH: True}


def _isbool(x):
    isinstance(x, bool)


def _isint(x):
    isinstance(x, int)


def _isstr(x):
    isinstance(x, str)


_VALIDATORS = {
    NMAX: _isint,
    PARALLEL: _isbool,
    CACHE: _isbool,
    FASTMATH: _isbool,
}

_SETTERS = {}


class set_options(object):
    """Set options for xarray in a controlled context.
    Currently supported options:
    - `NMAX` : max moment size
    You can use ``set_options`` either as a context manager:
    >>> with xr.set_options(use_tqdm=True, tqdm_min_len_calc=50):
    ...     c.xge.betaOmega()
    ...
    Or to set global options:
    >>> xr.set_options(tqdm_min_len_calc=50)
    """

    def __init__(self, **kwargs):
        self.old = {}
        for k, v in kwargs.items():
            if k not in OPTIONS:
                raise ValueError(
                    "argument name %r is not in the set of valid options %r"
                    % (k, set(OPTIONS))
                )
            if k in _VALIDATORS and not _VALIDATORS[k](v):
                raise ValueError(f"option {k!r} given an invalid value: {v!r}")
            self.old[k] = OPTIONS[k]
        self._apply_update(kwargs)

    def _apply_update(self, options_dict):
        for k, v in options_dict.items():
            if k in _SETTERS:
                _SETTERS[k](v)
        OPTIONS.update(options_dict)

    def __enter__(self):
        return

    def __exit__(self, type, value, traceback):
        self._apply_update(self.old)



from __future__ import absolute_import

from functools import lru_cache

import numpy as np
import xarray as xr
from numba import njit

# from .cached_decorators import gcached  # , cached_clear
from .options import OPTIONS


def myjit(func):
    """
    "my" jit function
    uses option inline='always', fastmath=True
    """
    return njit(inline="always", fastmath=OPTIONS["fastmath"], cache=OPTIONS["cache"])(
        func
    )


# from scipy.special import binom
# def factory_binomial(order):
#     irange = np.arange(order + 1)
#     bfac = np.array([binom(i, irange) for i in irange])
#     return bfac


def _binom(n, k):
    if n > k:
        return np.math.factorial(n) / (np.math.factorial(k) * np.math.factorial(n - k))
    elif n == k:
        return 1.0
    else:
        # n < k
        return 0.0


def factory_binomial(order, dtype=np.float):
    out = np.zeros((order + 1, order + 1), dtype=dtype)
    for n in range(order + 1):
        for k in range(order + 1):
            out[n, k] = _binom(n, k)

    return out


def _my_broadcast(x, shape, dtype=None, order=None):
    x = np.asarray(x, dtype=dtype, order=order)
    if x.shape != shape:
        x = np.broadcast(x, shape)
    return x


def _shape_insert_axis(shape, axis, new_size):
    """
    given shape, get new shape with size put in position axis
    """
    n = len(shape)

    axis = np.core.numeric.normalize_axis_index(axis, n + 1)
    # assert -(n+1) <= axis <= n
    # if axis < 0:
    #     axis = axis + n + 1

    # if axis < 0:
    #     axis += len(shape) + 1
    shape = list(shape)
    shape.insert(axis, new_size)
    return tuple(shape)


def _shape_reduce(shape, axis):
    """given input shape, give shape after reducing along axis"""
    shape = list(shape)
    shape.pop(axis)
    return tuple(shape)


def _axis_expand_broadcast(
    x,
    shape,
    axis,
    verify=True,
    expand=True,
    broadcast=True,
    roll=True,
    dtype=None,
    order=None,
):
    """
    broadcast x to shape.  If x is 1d, and shape is n-d, but len(x) is same
    as shape[axis], broadcast x across all dimensions
    """

    if verify is True:
        x = np.asarray(x, dtype=dtype, order=order)

    # if array, and 1d with size same as shape[axis]
    # broadcast from here
    if expand:
        if x.ndim == 1 and x.ndim != len(shape) and len(x) == shape[axis]:
            # reshape for broadcasting
            reshape = [1] * (len(shape) - 1)
            reshape = _shape_insert_axis(reshape, axis, -1)
            x = x.reshape(*reshape)

    if broadcast and x.shape != shape:
        x = np.broadcast_to(x, shape)
    if roll and axis != 0:
        x = np.moveaxis(x, axis, 0)
    return x


@lru_cache(maxsize=5)
def _cached_ones(shape, dtype=None):
    return np.ones(shape, dtype=dtype)


def _xr_wrap_like(da, x):
    """
    wrap x with xarray like da
    """
    x = np.asarray(x)
    assert x.shape == da.shape

    return xr.DataArray(
        x, dims=da.dims, coords=da.coords, name=da.name, indexes=da.indexes
    )


def _xr_order_like(template, *others):
    """
    given dimensions, order in same manner
    """

    if not isinstance(template, xr.DataArray):
        out = others

    else:
        dims = template.dims

        key_map = {dim: i for i, dim in enumerate(dims)}

        def key(x):
            key_map[x]

        out = []
        for other in others:
            if isinstance(other, xr.DataArray):
                # reorder
                order = sorted(other.dims, key=key)

                x = other.transpose(*order)
            else:
                x = other

            out.append(x)

    if len(out) == 1:
        out = out[0]

    return out



"""
Routines to perform resampling
"""
from __future__ import absolute_import

from numba import njit, prange

from .options import OPTIONS
from .pushers import (
    _push_datas_scale,
    _push_datas_scale_cov,
    _push_datas_scale_cov_vec,
    _push_datas_scale_vec,
    _push_vals_scale,
    _push_vals_scale_cov,
    _push_vals_scale_cov_vec,
    _push_vals_scale_vec,
)

# from functools import lru_cache
# import numpy as np


def jitter(parallel):
    return njit(fastmath=OPTIONS["fastmath"], cache=OPTIONS["cache"], parallel=parallel)


# NOTE: this is all due to closures not being cache-able with numba
# used to use the following
#
# @lru_cache(10)
# def _factory_resample(push_datas_scale, fastmath=True, parallel=False):
#     @njit(fastmath=fastmath, parallel=parallel)
#     def resample(data, freq, out):
#         nrep = freq.shape[0]
#         for irep in prange(nrep):
#             push_datas_scale(out[irep, ...], data, freq[irep, ...])

#     return resample

# @lru_cache(10)
# def _factory_resample_vals(push_vals_scale, fastmath=True, parallel=False):
#     @njit(fastmath=fastmath, parallel=parallel)
#     def resample(W, X, freq, out):
#         nrep = freq.shape[0]
#         for irep in prange(nrep):
#             push_vals_scale(out[irep, ...], W, X, freq[irep, ...])

#     return resample


# @lru_cache(10)
# def _factory_resample_vals_cov(push_vals_scale, fastmath=True, parallel=False):
#     @njit(fastmath=fastmath, parallel=parallel)
#     def resample(W, X, Y, freq, out):
#         nrep = freq.shape[0]
#         for irep in prange(nrep):
#             push_vals_scale(out[irep, ...], W, X, Y, freq[irep, ...])

#     return resample

######################################################################
# resample data
# mom/scalar
@jitter(parallel=False)
def _resample_data(data, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_datas_scale(out[irep, ...], data, freq[irep, ...])


@jitter(parallel=True)
def _resample_data_parallel(data, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_datas_scale(out[irep, ...], data, freq[irep, ...])


# mom/vector
@jitter(parallel=False)
def _resample_data_vec(data, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_datas_scale_vec(out[irep, ...], data, freq[irep, ...])


@jitter(parallel=True)
def _resample_data_vec_parallel(data, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_datas_scale_vec(out[irep, ...], data, freq[irep, ...])


# cov/vector
@jitter(parallel=False)
def _resample_data_cov(data, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_datas_scale_cov(out[irep, ...], data, freq[irep, ...])


@jitter(parallel=True)
def _resample_data_cov_parallel(data, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_datas_scale_cov(out[irep, ...], data, freq[irep, ...])


# cov/vector
@jitter(parallel=False)
def _resample_data_cov_vec(data, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_datas_scale_cov_vec(out[irep, ...], data, freq[irep, ...])


@jitter(parallel=True)
def _resample_data_cov_vec_parallel(data, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_datas_scale_cov_vec(out[irep, ...], data, freq[irep, ...])


_RESAMPLE_DATA_DICT = {
    # cov, vec, parallel
    (False, False, False): _resample_data,
    (False, False, True): _resample_data_parallel,
    (False, True, False): _resample_data_vec,
    (False, True, True): _resample_data_vec_parallel,
    (True, False, False): _resample_data_cov,
    (True, False, True): _resample_data_cov_parallel,
    (True, True, False): _resample_data_cov_vec,
    (True, True, True): _resample_data_cov_vec_parallel,
}


def factory_resample_data(cov, vec, parallel):
    return _RESAMPLE_DATA_DICT[cov, vec, parallel]


######################################################################
# resample values


# mom/scalar
@jitter(parallel=False)
def _resample_vals(W, X, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_vals_scale(out[irep, ...], W, X, freq[irep, ...])


@jitter(parallel=True)
def _resample_vals_parallel(W, X, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_vals_scale(out[irep, ...], W, X, freq[irep, ...])


# mom/vec
@jitter(parallel=False)
def _resample_vals_vec(W, X, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_vals_scale_vec(out[irep, ...], W, X, freq[irep, ...])


@jitter(parallel=True)
def _resample_vals_vec_parallel(W, X, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_vals_scale_vec(out[irep, ...], W, X, freq[irep, ...])


# cov/scalar
@jitter(parallel=False)
def _resample_vals_cov(W, X, Y, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_vals_scale_cov(out[irep, ...], W, X, Y, freq[irep, ...])


@jitter(parallel=True)
def _resample_vals_cov_parallel(W, X, Y, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_vals_scale_cov(out[irep, ...], W, X, Y, freq[irep, ...])


# cov/vec
@jitter(parallel=False)
def _resample_vals_cov_vec(W, X, Y, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_vals_scale_cov_vec(out[irep, ...], W, X, Y, freq[irep, ...])


@jitter(parallel=True)
def _resample_vals_cov_vec_parallel(W, X, Y, freq, out):
    nrep = freq.shape[0]
    for irep in prange(nrep):
        _push_vals_scale_cov_vec(out[irep, ...], W, X, Y, freq[irep, ...])


_RESAMPLE_VALS_DICT = {
    # cov, vec, parallel
    (False, False, False): _resample_vals,
    (False, False, True): _resample_vals_parallel,
    (False, True, False): _resample_vals_vec,
    (False, True, True): _resample_vals_vec_parallel,
    (True, False, False): _resample_vals_cov,
    (True, False, True): _resample_vals_cov_parallel,
    (True, True, False): _resample_vals_cov_vec,
    (True, True, True): _resample_vals_cov_vec_parallel,
}


def factory_resample_vals(cov, vec, parallel):
    return _RESAMPLE_VALS_DICT[cov, vec, parallel]



"""
Container for multiple (similar) CentralMoments objects
"""
from __future__ import absolute_import

import numpy as np

from .cached_decorators import cached_clear, gcached

# from .central import CentralMoments, CentralMomentsCov
# from .resample import randsamp_freq, resample_data, resample_vals
# from .utils import (
#     _axis_expand_broadcast,
#     _cached_ones,
#     _my_broadcast,
#     _shape_insert_axis,
# )


def weighted_var(x, w, axis=None, axis_sum=None, unbiased=True, **kwargs):
    """
    return the weighted variance over x with weight w

    v = sum(w)**2/(sum(w)**2 - sum(w**2)) * sum(w * (x-mu)**2 )

    Parameters
    ----------
    x : array
        values to consider

    w : array
        weights

    axis : axis to average over

    axis_sum : axis to sum over for  w,w**2

    unbiased : bool (default True)
    if True, then apply unbiased norm (like ddof=1)
    else, apply biased norm (like ddof=0)


    **kwargs : arguments to np.average

    Returns
    -------
    Ave : weighted average
        shape x with `axis` removed

    Var : weighted variance
        shape x with `axis` removed
    """

    if axis_sum is None:
        axis_sum = axis

    m1 = np.average(x, weights=w, axis=axis, **kwargs)
    m2 = np.average((x - m1) ** 2, weights=w, axis=axis, **kwargs)

    if unbiased:
        w1 = w.sum(axis=axis_sum)
        w2 = (w * w).sum(axis=axis_sum)
        m2 *= w1 * w1 / (w1 * w1 - w2)
    return m1, m2


class StatsArray(object):
    """
    Collection of Accumulator objects
    """

    def __init__(self, obj):
        """
        Parameters
        ----------
        obj : CentralMoments
        """
        self._accum = obj
        self.zero()

        # if isinstance(moments, int):
        #     moments = (moments,)

        # assert isinstance(moments, tuple)

        # if child is None:
        #     if len(moments) == 1:
        #         child = CentralMoments
        #     else:
        #         child = CentralMomentsCov

        # self._child = child
        # self._accum = child(shape=shape, moments=moments, dtype=dtype)
        # self.zero()

    @classmethod
    def zeros(cls, *args, **kwargs):
        """create empty array object with baseclass CentralMoments"""
        pass

    @classmethod
    def xzeros(cls, *args, **kwargs):
        """create empty array object with baseclass xCentralMoments"""
        pass

    @property
    def accum(self):
        return self._accum

    @property
    def mom(self):
        return self._accum.moments

    @property
    def dtype(self):
        return self._accum.dtype

    @property
    def values(self):
        return self._values

    @values.setter
    @cached_clear()
    def values(self, values):
        if not isinstance(values, list):
            raise ValueError("trying to set list to non-list value")
        self._values = values

    @gcached()
    def data(self):
        return np.array(self._values)

    def new_like(self):
        return self.__class__(
            shape=self.accum.shape,
            child=self._child,
            dtype=self.dtype,
            moments=self.accum.moments,
        )

    def __len__(self):
        return len(self._values)

    def __getitem__(self, idx):
        new = self.new_like()
        try:
            y = self._values[idx]
        except Exception:
            y = list(self.data[idx])
        if not isinstance(y, list):
            y = [y]

        new._values = y
        return new

    def to_stats(self, indices=None, copy=True):
        data = self.data
        if indices is not None:
            data = np.take(data, indices, axis=0)
        elif copy:
            data = data.copy()
        return self._child.from_data(data, moments=self.moments, copy=False)

    def resample(self, indices):
        """
        axis = axis of indices to average over
        """
        return self.to_stats(indices)

    def resample_and_reduce(self, freq, **kwargs):
        """
        for bootstrapping
        """
        data = self.data
        data_new = data.resample_data(data, freq, moments=self.moments, **kwargs)
        return self.__class__.from_datas(
            data_new, shape=self._accum.shape, child=self._child, moments=self.moments
        )

    def zero(self):
        self.values = []
        self.accum.zero()

    @cached_clear()
    def append(self, data):
        assert data.shape == self._accum.shape_data
        self._values.append(data)

    @cached_clear()
    def push_stat(self, a, v=0.0, w=1.0):
        s = self._child.from_stat(a=a, v=v, w=w)
        self._values.append(s.data)

    @cached_clear()
    def push_stats(self, a, v=None, w=None):
        if v is None:
            v = np.zeros_like(a)
        if w is None:
            w = np.ones_like(a)
        for (ww, aa, vv) in zip(w, a, v):
            self.push_stat(a=aa, v=vv, w=ww)

    @cached_clear()
    def push_data(self, data, copy=False):
        assert data.shape == self._accum.shape_data
        if copy:
            data = data.copy()
        self._values.append(data)

    @cached_clear()
    def push_datas(self, datas, axis=0):
        if axis != 0:
            datas = np.moveaxis(datas, axis, 0)
        assert datas.shape[1:] == self._accum.shape_data
        for data in datas:
            self._values.append(data)

    @gcached()
    def _weight_index(self):
        return (slice(None),) + self._accum._weight_index

    @gcached(prop=False)
    def _single_index(self, val):
        return (slice(None),) + self._accum._single_index(val)

    def weight(self):
        return self.data[self._weight_index]

    def mean(self):
        return self.data[self._single_index(1)]

    def var(self):
        return self.data[self._single_index(2)]

    @property
    def data_last(self):
        return self._values[-1]

    def mean_last(self):
        return self.data_last[self.accum._single_index(1)]

    def var_last(self):
        return self.data_last[self.accum._single_index(2)]

    def std_last(self):
        return np.sqrt(self.var_last())

    def weight_last(self):
        return self.data_last[self.accum._weight_index]

    def get_stat(self, stat_name="mean", *args, **kwargs):
        return getattr(self, stat_name)(*args, **kwargs)

    @classmethod
    def from_datas(cls, datas, moments, axis=0, shape=None, child=None, dtype=np.float):
        if isinstance(moments, int):
            moments = (moments,)

        if axis != 0:
            datas = np.moveaxis(datas, axis, 0)

            shape = datas.shape[1 : -len(moments)]

        new = cls(child=child, shape=shape, moments=moments, dtype=dtype)

        new.values = list(datas)
        return new

    @classmethod
    def from_accum(cls, accum, axis=None):
        """
        create StatsArray from CentralMoments object

        if accum object is a scalar object, or a vector object with no specified axis,
        then create a StatsArray object with accum as the sole elements

        if accum object is a vector object with a specified axis, create a
        StatsArray object
        with elements along this axis
        """

        ndim = len(accum.moments)
        if ndim == 0:
            axis = None
        if axis is None:
            # create single object
            new = cls(
                moments=accum.moments,
                shape=accum.shape,
                child=type(accum),
                dtype=accum.dtype,
            )

        else:
            new = cls.from_datas(
                accum.data,
                moments=accum.moments,
                axis=axis,
                shape=None,
                child=type(accum),
                dtype=accum.dtype,
            )

        return new

    @gcached()
    def cumdata(self):
        cumdata = np.zeros((len(self),) + self.accum.shape_data)
        self._accum.zero()
        for i, data in enumerate(self.values):
            self._accum.push_data(data)
            cumdata[i, ...] = self._accum.data
        return cumdata

    def cummean(self):
        return self.cumdata[self._single_index(1)]

    def cumvar(self):
        return self.cumdata[self._single_index(2)]

    def cumstd(self):
        return np.sqrt(self.cumvar())

    def cumweight(self):
        return self.cumdata[self._weight_index]

    @property
    def cumdata_last(self):
        return self.cumdata[-1, ...]

    def cummean_last(self):
        return self.cumdata_last[self.accum._single_index(1)]

    def cumvar_last(self):
        return self.cumdata_last[self.accum._single_index(2)]

    def cumstd_last(self):
        return np.sqrt(self.cumvar_last())

    def cumweight_last(self):
        return self.cumdata_last[self.accum._weight_index]

    @gcached()
    def stats_list(self):
        """
        list of stats objects
        """
        return [
            self._child.from_data(
                data=data,
                shape=self.accum.shape,
                moments=self.moments,
                dtype=self.dtype,
                copy=False,
            )
            for data in self.values
        ]

    def block(self, block_size=None):
        """
        create a new stats array object from block averaging this one
        """
        new = self.new_like()
        new.values = self.blockdata(block_size)
        return new

    @gcached(prop=False)
    def blockdata(self, block_size):
        blockdata = []

        n = len(self)
        if block_size is None:
            block_size = n
        if block_size > n:
            block_size = n

        for lb in range(0, len(self), block_size):
            ub = lb + block_size
            if ub > n:
                break
            self._accum.zero()
            datas = self.data[lb:ub, ...]
            self._accum.push_datas(datas)
            blockdata.append(self._accum.data.copy())
        return blockdata

    def blockweight(self, block_size=None):
        return self.blockdata(block_size)[self._weight_index]

    def blockmean(self, block_size=None):
        return self.blockdata(block_size)[self._single_index(1)]

    def blockvar(self, block_size=None):
        return self.blockdata(block_size)[self._single_index(2)]

    def val_SEM(self, x, weighted, unbiased, norm):
        """
        find the standard error in the mean (SEM) of a value

        Parameters
        ----------
        x : array
            array (self.mean(), etc) to consider

        weighted : bool
            if True, use `weighted_var`
            if False, use `np.var`

        unbiased : bool
            if True, use unbiased stats (e.g., ddof=1 for np.var)
            if False, use biased stats (e.g., ddof=0 for np.var)

        norm : bool
            if True, scale var by x.shape[0], i.e., number of samples

        Returns
        -------
        sem : standard error in mean
        """
        if weighted:
            v = weighted_var(x, w=self.weight(), axis=0, unbiased=unbiased)[-1]
        else:
            if unbiased:
                ddof = 1
            else:
                ddof = 0

            v = np.var(x, ddof=ddof, axis=0)
        if norm:
            v = v / x.shape[0]

        return np.sqrt(v)

    def mean_SEM(self, weighted=True, unbiased=True, norm=True):
        """self.val_SEM with x=self.mean()"""
        return self.val_SEM(self.mean(), weighted, unbiased, norm)

    def __repr__(self):
        return "nsample: {}".format(len(self))

    def to_xarray(
        self,
        rec_dim="rec",
        meta_dims=None,
        mom_dims=None,
        rec_coords=None,
        meta_coords=None,
        mom_coords=None,
        **kwargs,
    ):
        import xarray as xr

        if meta_dims is None:
            meta_dims = ["dim_{}".format(i) for i in range(len(self.accum.shape))]
        else:
            meta_dims = list(meta_dims)
        assert len(meta_dims) == len(self.accum.shape)

        if mom_dims is None:
            mom_dims = ["mom_{}".format(i) for i in range(len(self.moments))]

        if isinstance(mom_dims, str):
            mom_dims = [mom_dims]

        assert len(mom_dims) == len(self.accum.moments)

        dims = [rec_dim] + meta_dims + mom_dims

        coords = {}
        coords.update(rec_coords or {})
        coords.update(meta_coords or {})
        coords.update(mom_coords or {})
        return xr.DataArray(self.data, dims=dims, coords=coords, **kwargs)

    @classmethod
    def from_xarray(
        cls,
        data,
        rec_dim="rec",
        meta_dims=None,
        mom_dims=None,
        shape=None,
        moments=None,
        child=None,
        dtype=None,
    ):
        pass

        if mom_dims is None:
            # try to infer moment dimensions
            mom_dims = []
            for k in sorted(data.dims):
                if "mom_" in k:
                    mom_dims.append(k)

        if isinstance(mom_dims, str):
            mom_dims = [mom_dims]

        if moments is None:
            # infer moments
            moments = []
            for k in mom_dims:
                moments.append(len(data[k]) - 1)
            moments = tuple(moments)

        assert len(moments) == len(mom_dims)

        order = [rec_dim]
        if meta_dims is not None:
            if isinstance(meta_dims, str):
                meta_dims = [meta_dims]
            assert data.ndim == 1 + len(mom_dims) + len(meta_dims)
            order += meta_dims
        else:
            order += [...]

        order += mom_dims

        data = data.transpose(*order)

        return cls.from_datas(
            datas=data, moments=moments, axis=0, shape=shape, child=child, dtype=dtype
        )



"""
Set of routines to call most cases to pre-compile numba functions
"""

from __future__ import absolute_import

import numpy as np

from .central import CentralMoments


def compile_numba_funcs():
    dtype = np.float
    for val_shape in [(), (1,), (1, 1)]:
        for mom in [(1,), (1, 1)]:
            s = CentralMoments.zeros(val_shape=val_shape, mom=mom)

            val = np.empty(val_shape, dtype=dtype)
            vals = np.empty((1,) + val_shape, dtype=dtype)

            if len(mom) == 2:
                val = (val, val)
                vals = (vals, vals)

            data = np.empty(s.shape, dtype=dtype)
            datas = np.empty((1,) + s.shape, dtype=dtype)

            s.push_val(val)
            s.push_vals(vals)

            s.push_data(data)
            s.push_datas(datas)

            if len(val_shape) > 1:
                for parallel in [False, True]:
                    s.resample_and_reduce(nrep=2, parallel=parallel)



"""
Central moments/comoments routines
"""
from __future__ import absolute_import

import numpy as np
from numpy.core.numeric import normalize_axis_index  # , normalize_axis_tuple

from . import convert
from .cached_decorators import gcached
from .pushers import factory_pushers
from .resample import randsamp_freq, resample_data, resample_vals
from .utils import _axis_expand_broadcast  # _cached_ones,; _my_broadcast,
from .utils import _shape_insert_axis, _shape_reduce


###############################################################################
# central mom/comoments routines
###############################################################################
def _central_moments(
    x, mom, w=None, axis=0, last=True, dtype=None, order=None, out=None
):
    """calculate central mom along axis"""

    if isinstance(mom, tuple):
        mom = mom[0]

    x = np.asarray(x, dtype=dtype, order=order)
    if dtype is None:
        dtype = x.dtype

    if w is None:
        w = np.ones_like(x)
    else:
        w = _axis_expand_broadcast(
            w, x.shape, axis, roll=False, dtype=dtype, order=order
        )

    # if axis < 0:
    #     axis += x.ndim
    if axis != 0:
        x = np.moveaxis(x, axis, 0)
        w = np.moveaxis(w, axis, 0)

    shape = (mom + 1,) + x.shape[1:]
    if out is None:
        out = np.empty(shape, dtype=dtype)
    else:
        if out.shape != shape:
            # try rolling
            out = np.moveaxis(out, -1, 0)
        assert out.shape == shape

    wsum = w.sum(axis=0)
    wsum_inv = 1.0 / wsum
    xave = np.einsum("r...,r...->...", w, x) * wsum_inv

    shape = (-1,) + (1,) * (x.ndim)
    p = np.arange(2, mom + 1).reshape(*shape)

    dx = (x[None, ...] - xave) ** p

    out[0, ...] = wsum
    out[1, ...] = xave
    out[2:, ...] = np.einsum("r..., mr...->m...", w, dx) * wsum_inv

    if last:
        out = np.moveaxis(out, 0, -1)
    return out


def _central_comoments(
    x,
    mom,
    w=None,
    axis=0,
    last=True,
    broadcast=False,
    dtype=None,
    order=None,
    out=None,
):
    """calculate central co-mom (covariance, etc) along axis"""

    if isinstance(mom, int):
        mom = (mom,) * 2

    mom = tuple(mom)
    assert len(mom) == 2

    # change x to tuple of inputs
    assert isinstance(x, tuple) and len(x) == 2
    x, y = x

    x = np.asarray(x, dtype=dtype, order=order)
    if dtype is None:
        dtype = x.dtype

    y = _axis_expand_broadcast(
        y,
        x.shape,
        axis,
        roll=False,
        broadcast=broadcast,
        expand=broadcast,
        dtype=dtype,
        order=order,
    )

    if w is None:
        w = np.ones_like(x)
    else:
        w = _axis_expand_broadcast(
            w, x.shape, axis, roll=False, dtype=dtype, order=order
        )

    assert w.shape == x.shape
    assert y.shape == x.shape

    # if axis < 0:
    #     axis += x.ndim
    if axis != 0:
        x = np.moveaxis(x, axis, 0)
        y = np.moveaxis(y, axis, 0)
        w = np.moveaxis(w, axis, 0)

    shape = tuple(x + 1 for x in mom) + x.shape[1:]
    if out is None:
        out = np.empty(shape, dtype=dtype)
    else:
        if out.shape != shape:
            # try moving axis
            out = np.moveaxis(out, [-2, -1], [0, 1])
        assert out.shape == shape

    wsum = w.sum(axis=0)
    wsum_inv = 1.0 / wsum

    xave = np.einsum("r...,r...->...", w, x) * wsum_inv
    yave = np.einsum("r...,r...->...", w, y) * wsum_inv

    shape = (-1,) + (1,) * (x.ndim)
    p0 = np.arange(0, mom[0] + 1).reshape(*shape)
    p1 = np.arange(0, mom[1] + 1).reshape(*shape)

    dx = (x[None, ...] - xave) ** p0
    dy = (y[None, ...] - yave) ** p1

    out[...] = np.einsum("r...,ir...,jr...->ij...", w, dx, dy) * wsum_inv

    out[0, 0, ...] = wsum
    out[1, 0, ...] = xave
    out[0, 1, ...] = yave

    if last:
        out = np.moveaxis(out, [0, 1], [-2, -1])
    return out


def central_moments(
    x, mom, w=None, axis=0, last=True, dtype=None, order=None, out=None, broadcast=False
):
    """
    calculate central moments or comoments along axis

    Parameters
    ----------
    x : array-like or tuple of array-like
        if calculating moments, then this is the input array.
        if calculating comoments, then pass in tuple of values of form (x, y)
    mom : int or tuple
        number of moments to calculate.  If tuple, then this specifies that
        comoments will be calculated.
    w : array-like, optional
        Weights. If passed, should be able to broadcast to `x`. An exception is if
        w is a 1d array with len(w) == x.shape[axis]. In this case,
        w will be reshaped and broadcast against x
    axis : int, default=0
        axis to reduce along
    last : bool, aaefault=True
        if True, put mom as last dimension.
        Otherwise, mom will be in first dimension
    dtype, order : options to np.asarray
    broadcast : bool, default=False
        if True and calculating comoments, then the x[1] will be broadcast against x[0].
    out : array
        if present, use this for output data
        Needs to have shape of either (mom,) + shape or shape + (mom,)
        where shape is the shape of x with axis removed

    Returns
    -------
    output : array
        array of shape=shape + mom_shape or mom_shape + shape depending on
        value of `last`, where `shape` is the shape of `x` with axis removed,
        i.e., shape=x.shape[:axis] + x.shape[axis+1:], and `mom_shape` is the shape of
        the moment part, either (mom+1,) or (mom0+1, mom1+1).  Assuming `last is True`,
        output[...,0] is the total weight (or count), output[...,1] is the mean
        value of x, output[...,n] with n>1 is the nth central moment.
    """

    if isinstance(mom, int):
        mom = (mom,)

    kws = dict(
        x=x, mom=mom, w=w, axis=axis, last=last, dtype=dtype, order=order, out=out
    )
    if len(mom) == 1:
        func = _central_moments
    else:
        func = _central_comoments
        kws["broadcast"] = broadcast

    return func(**kws)


###############################################################################
# Classes
###############################################################################
class CentralMoments(object):
    """
    Base class for moments accumulation
    """

    __slots__ = (
        "_mom_ndim",
        "_cache",
        "_data",
        "_data_flat",
        "_push",
    )

    def __init__(self, data, mom_ndim=1):
        """
        Parameters
        ----------
        data : numpy data array
            data should have the shape `val_shape + mom_shape`
            where `val_shape` is the shape of a single observation,
            and mom_shape is the shape of the moments
        mom_ndim : int, {1, 2}
            number of dimensions of moments.
            * 1 : central moments of single variable
            * 2 : central comoments of two variables
        **kws : dict
            optional arguments to be used in subclasses
        """

        if mom_ndim not in (1, 2):
            raise ValueError(
                "mom_ndim must be either 1 (for central moments)"
                "or 2 (for central comoments)"
            )
        self._mom_ndim = mom_ndim

        if data.ndim < self.mom_ndim:
            raise ValueError("not enough dimensions in data")

        self._data = data

        # check moments make sense
        # skip this as it comes from array shape
        # if not all([x > 0 for x in self.mom]):
        #     raise ValueError("All moments must positive")

        self._data_flat = self._data.reshape(self.shape_flat)

        # setup pushers
        vec = len(self.val_shape) > 0
        cov = self.mom_ndim == 2
        self._push = factory_pushers(cov=cov, vec=vec)

    @property
    def values(self):
        """accessor to underlying central moments data"""
        return self._data

    @property
    def data(self):
        """accessor to numpy underlying data

        By convention data has the following meaning for the moments indexes

        * `data[...,i=0,j=0]`, weights
        * `data[...,i=1,j=0]]`, if only one moment indice is one and all
        others zero, then this is the average value of the variable with unit index.

        * all other cases, the central moments `<(x0-<x0>)**i0 * (x1 - <x1>)**i1 * ...>`
        """
        return self._data

    @property
    def shape(self):
        """self.data.shape"""
        return self._data.shape

    @property
    def ndim(self):
        """self.data.ndim"""
        return self._data.ndim

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def mom(self):
        """number of moments for each variable"""
        return tuple(x - 1 for x in self.mom_shape)

    @property
    def mom_shape(self):
        """shape of moments part"""
        return self._data.shape[-self.mom_ndim :]

    @property
    def mom_ndim(self):
        """length of moment part
        if `mom_ndim` == 1, then single variable moments
        if `mom_ndim` == 2, then co-moments
        """
        return self._mom_ndim

    @property
    def val_shape(self):
        """shape, less moments dimensions"""
        return self._data.shape[: -self.mom_ndim]

    @property
    def val_ndim(self):
        return len(self.val_shape)

    @property
    def val_shape_flat(self):
        """shape of values part flattened"""
        if self.val_shape == ():
            return ()
        else:
            return (np.prod(self.val_shape),)

    @property
    def shape_flat(self):
        return self.val_shape_flat + self.mom_shape

    @property
    def mom_shape_var(self):
        """shape of moment part of variance"""
        return tuple(x - 1 for x in self.mom)

    # variance shape
    @property
    def shape_var(self):
        """total variance shape"""
        return self.val_shape + self.mom_shape_var

    @property
    def shape_flat_var(self):
        return self.val_shape_flat + self.mom_shape_var

    # I think this is for pickling
    # probably don't need it anymore
    # def __setstate__(self, state):
    #     self.__dict__ = state
    #     # make sure datar points to data
    #     self._data_flat = self._data.reshape(self.shape_flat)

    # not sure I want to actually implements this
    # could lead to all sorts of issues applying
    # ufuncs to underlying data
    # def __array_wrap__(self, obj, context=None):
    #     return self, obj, context

    def __repr__(self):
        s = "<CentralMoments(val_shape={}, mom={})>\n".format(self.val_shape, self.mom)
        return s + repr(self.values)

    def __array__(self, dtype=None):
        return np.asarray(self._data, dtype=dtype)

    ###########################################################################
    # SECTION: top level creation/copy/new
    ###########################################################################
    def new_like(
        self, data=None, verify=False, check=False, copy=False, copy_kws=None, **kws
    ):
        """create new object like self, with new data

        Parameters
        ----------
        data : array-like, optional
            data for new object
        verify : bool, default=False
            if True, pass data through np.asarray
        check : bool, default=True
            if True, then check that data has same total shape as self
        copy : bool, default=False
            if True, perform copy of data
        copy_kws : dict, optional
            key-word arguments to `self.data.copy`
        **kws : extra arguments
            arguments to `type(self).__init__`
        """

        if data is None:
            data = np.zeros_like(self._data)
        else:
            if verify:
                data = np.asarray(data, dtype=self.dtype)

            if check:
                assert data.shape == self.shape

            if copy:
                if copy_kws is None:
                    copy_kws = {}
                data = data.copy(**copy_kws)
        return type(self)(data=data, mom_ndim=self.mom_ndim, **kws)

    def zeros_like(self, zeros_kws=None, **kws):
        """create new object empty object like self"""
        if zeros_kws is None:
            zeros_kws = {}
        return self.new_like(data=np.zeros_like(self._data, **zeros_kws), **kws)

    def copy(self, **copy_kws):
        """create a new object with copy of data"""
        return self.new_like(
            data=self.values, verify=False, check=False, copy=True, copy_kws=copy_kws
        )

    @classmethod
    def zeros(
        cls,
        mom=None,
        val_shape=None,
        mom_ndim=None,
        shape=None,
        dtype=None,
        zeros_kws=None,
        **kws,
    ):
        """create a new base object

        Parameters
        ----------
        mom : int or tuple
            moments.
            if integer, or length one tuple, then moments of single variable.
            if tuple of length 2, then comoments of two variables.
        val_shape : tuple, optional
            shape of values, excluding moments.  For example, if considering the average
            of observations `x`, then `val_shape = x.shape`
            if not passed, then assume val_shape = ()
        shape : tuple, optional
            if passed, create object with this total shape
        mom_ndim : int {1, 2}, optional
            number of variables.
            if pass `shape`, then must pass mom_ndim
        dtype : nunpy dtype, default=float
        zeros_kws : dict
        extra arguments to `np.zeros`
        kws : dict
            extra arguments to `__init__`

        Returns
        -------
        object : instance of class `cls`

        Notes
        -----
        the resulting total shape of data is shape + (mom + 1)
        """

        if zeros_kws is None:
            zeros_kws = {}

        if shape is None:
            assert mom is not None
            if isinstance(mom, int):
                mom = (mom,)
            if mom_ndim is None:
                mom_ndim = len(mom)
            assert len(mom) == mom_ndim

            if val_shape is None:
                val_shape = ()
            elif isinstance(val_shape, int):
                val_shape = (val_shape,)
            shape = val_shape + tuple(x + 1 for x in mom)

        else:
            assert mom_ndim is not None

        if dtype is None:
            dtype = np.float

        if zeros_kws is None:
            zeros_kws = {}
        data = np.zeros(shape=shape, dtype=dtype, **zeros_kws)

        return cls(data=data, mom_ndim=mom_ndim, **kws)

    ###########################################################################
    # SECTION: Access to underlying statistics
    ###########################################################################
    @gcached()
    def _weight_index(self):
        index = [0] * len(self.mom)
        if self.val_ndim > 0:
            index = [...] + index
        return tuple(index)

    @gcached(prop=False)
    def _single_index(self, val):
        # index with things like data[..., 1,0] data[..., 0,1]
        # index = (...,[1,0],[0,1])
        dims = len(self.mom)
        if dims == 1:
            index = [val]
        else:
            # this is a bit more complicated
            index = [[0] * dims for _ in range(dims)]
            for i in range(dims):
                index[i][i] = val

        if self.val_ndim > 0:
            index = [...] + index

        return tuple(index)

    def weight(self):
        return self.values[self._weight_index]

    def mean(self):
        return self.values[self._single_index(1)]

    def var(self):
        return self.values[self._single_index(2)]

    def std(self):
        return np.sqrt(self.var())

    def cmom(self):
        """central moments
        `cmom[..., i0, i1] = < (x0 - <x0>)**i0 * (x1 - <x1>)**i1>`
        """
        out = self.data.copy()
        # zeroth central moment
        out[self._weight_index] = 1
        # first central moment
        out[self._single_index(1)] = 0
        return out

    # convert to/from raw moments
    def to_raw(self):
        """convert central moments to raw moments

        raw[...,i, j] = weight,           i = j = 0
                      = <x0**i * x1**j>,  otherwise
        """
        if self.mom_ndim == 1:
            func = convert.to_raw_moments
        elif self.mom_ndim == 2:
            func = convert.to_raw_comoments
        return func(self.data)

    def rmom(self):
        """raw moments
        rmom[..., i, j] = <x0 ** i * x1 ** j>
        """
        out = self.to_raw()
        out[self._weight_index] = 1
        return out

    ###########################################################################
    # SECTION: pushing routines
    ###########################################################################
    def _asarray(self, val):
        return np.asarray(val, dtype=self.dtype)

    # @property
    # def _unit_weight(self):
    #     """internally cached unit weight"""
    #     return _cached_ones(self.shape, dtype=self.dtype)

    def _verify_value(
        self,
        x,
        target=None,
        axis=None,
        broadcast=False,
        expand=False,
        shape_flat=None,
        other=None,
    ):
        """
        verify input values

        Parameters
        ----------
        x : array
        target : tuple or array
            If tuple, this is the target shape to be used to Make target.
            If array, this is the target array
        Optinal target that has already been rolled.  If this is passed, and
        x will be broadcast/expanded, can expand to this shape without the need
        to reorder,
        """

        x = self._asarray(x)

        if isinstance(target, str):
            if target == "val":
                target = self.val_shape
            elif target == "vals":
                target = _shape_insert_axis(self.val_shape, axis, x.shape[axis])
            elif target == "data":
                target = self.shape
            elif target == "datas":
                # make sure axis in limits
                axis = normalize_axis_index(axis, self.val_ndim + 1)
                # if axis < 0:
                #     axis += self.ndim - self.mom_ndim
                target = _shape_insert_axis(self.shape, axis, x.shape[axis])
            elif target == "var":
                target = self.shape_var
            elif target == "vars":
                target = _shape_insert_axis(self.shape_var, axis, other.shape[axis])

        if isinstance(target, tuple):
            target_shape = target
            target_output = x

        else:
            target_shape = target.shape
            target_output = None

        x = _axis_expand_broadcast(
            x,
            target_shape,
            axis,
            verify=False,
            expand=expand,
            broadcast=broadcast,
            dtype=self.dtype,
            roll=False,
        )

        # check shape:
        assert (
            x.shape == target_shape
        ), f"x.shape = {x.shape} not equal target_shape={target_shape}"

        # move axis
        if axis is not None:
            if axis != 0:
                x = np.moveaxis(x, axis, 0)
            nrec = (x.shape[0],)
        else:
            nrec = ()

        if shape_flat is not None:
            x = x.reshape(nrec + shape_flat)

        if x.ndim == 0:
            x = x[()]

        if target_output is None:
            return x
        else:
            return x, target_output

    def check_weight(self, w, target):
        if w is None:
            w = 1.0
        return self._verify_value(
            w,
            target=target,
            axis=None,
            broadcast=True,
            expand=True,
            shape_flat=self.val_shape_flat,
        )

    def check_weights(self, w, target, axis=0):
        if w is None:
            w = 1.0
        return self._verify_value(
            w,
            target=target,
            axis=axis,
            broadcast=True,
            expand=True,
            shape_flat=self.val_shape_flat,
        )

    def check_val(self, x, target, broadcast=False):
        return self._verify_value(
            x,
            target=target,
            broadcast=broadcast,
            expand=False,
            shape_flat=self.val_shape_flat,
        )

    def check_vals(self, x, target, axis=0, broadcast=False):
        return self._verify_value(
            x,
            target=target,
            axis=axis,
            broadcast=broadcast,
            expand=broadcast,
            shape_flat=self.val_shape_flat,
        )

    def check_var(self, v, broadcast=False):
        return self._verify_value(
            v,
            target="var",  # self.shape_var,
            broadcast=broadcast,
            expand=False,
            shape_flat=self.shape_flat_var,
        )[0]

    def check_vars(self, v, target, axis=0, broadcast=False):
        return self._verify_value(
            v,
            target="vars",
            axis=axis,
            broadcast=broadcast,
            expand=broadcast,
            shape_flat=self.shape_flat_var,
            other=target,
        )[0]

    def check_data(self, data):
        return self._verify_value(data, target="data", shape_flat=self.shape_flat)[0]

    def check_datas(self, datas, axis=0):
        return self._verify_value(
            datas, target="datas", axis=axis, shape_flat=self.shape_flat
        )[0]

    def fill(self, value=0):
        """fill data with value"""
        self._data.fill(value)
        return self

    def zero(self):
        """zero out underlying data"""
        return self.fill(value=0)

    def push_data(self, data):
        """push `data` to moments

        Parameters
        ----------
        data : array-like, shape=self.shape
            array storing moment information
        Returns
        -------
        self

        See Also
        --------
        `self.data`
        """
        data = self.check_data(data)
        self._push.data(self._data_flat, data)
        return self

    def push_datas(self, datas, axis=0):
        """push and reduce multiple average central moments

        Parameters
        ----------
        datas : array-like
            this should have shape like `(nrec,) + self.shape`
            if `axis=0`, where `nrec` is the number of data objects to sum.
        axis : int, default=0
            axis to reduce along

        Returns
        -------
        self
        """
        datas = self.check_datas(datas, axis)
        self._push.datas(self._data_flat, datas)
        return self

    def push_val(self, x, w=None, broadcast=False):
        """dd single sample to central moments

        Parameters
        ----------
        x : array-like or tuple of arrays
            if `self.mom_ndim` == 1, then this is the value to consider
            if `self.mom_ndim` == 2, then x = (x0, x1)
            `x.shape == self.val_shape`

        w : int, float, array-like, optional
            optional weight of each sample
        broadcast : bool, default = False
            If true, do smart broadcasting for `x[1:]`

        Returns
        -------
        self
        """

        if self.mom_ndim == 1:
            ys = ()
        else:
            assert len(x) == self.mom_ndim
            x, *ys = x

        xr, target = self.check_val(x, "val")
        yr = tuple(self.check_val(y, target=target, broadcast=broadcast) for y in ys)
        wr = self.check_weight(w, target)
        self._push.val(self._data_flat, *((wr, xr) + yr))
        return self

    def push_vals(self, x, w=None, axis=0, broadcast=False):
        """
        add multiple samples to central moments

        Parameters
        ----------
        x : array-like or tuple of arrays
            if `self.mom_ndim` == 1, then this is the value to consider
            if `self.mom_ndim` == 2, then x = (x0, x1)
            `x.shape[:axis] + x.shape[axis+1:] == self.val_shape`

        w : int, float, array-like, optional
            optional weight of each sample
        axis : int, default=0
            axis to reduce along
        broadcast : bool, default = False
            If true, do smart broadcasting for `x[1:]`
        """
        if self.mom_ndim == 1:
            ys = ()
        else:
            assert len(x) == self.mom_ndim
            x, *ys = x

        xr, target = self.check_vals(x, axis=axis, target="vals")
        yr = tuple(
            self.check_vals(y, target=target, axis=axis, broadcast=broadcast)
            for y in ys
        )
        wr = self.check_weights(w, target=target, axis=axis)
        self._push.vals(self._data_flat, *((wr, xr) + yr))
        return self

    ###########################################################################
    # SECTION: Operators
    ###########################################################################
    def _check_other(self, b):
        """check other object"""
        assert type(self) == type(b)
        assert self.mom_ndim == b.mom_ndim
        assert self.shape == b.shape

    def __iadd__(self, b):
        self._check_other(b)
        # self.push_data(b.data)
        # return self
        return self.push_data(b.data)

    def __add__(self, b):
        self._check_other(b)
        # new = self.copy()
        # new.push_data(b.data)
        # return new
        return self.copy().push_data(b.data)

    def __isub__(self, b):
        # NOTE: consider implementint push_data_scale routine to make this cleaner
        self._check_other(b)
        assert np.all(self.weight() >= b.weight())
        data = b.data.copy()
        data[self._weight_index] *= -1
        # self.push_data(data)
        # return self
        return self.push_data(data)

    def __sub__(self, b):
        self._check_other(b)
        assert np.all(self.weight() >= b.weight())
        new = b.copy()
        new._data[self._weight_index] *= -1
        # new.push_data(self.data)
        # return new
        return new.push_data(self.data)

    def __mul__(self, scale):
        """
        new object with weights scaled by scale
        """
        scale = float(scale)
        new = self.copy()
        new._data[self._weight_index] *= scale
        return new

    def __imul__(self, scale):
        scale = float(scale)
        self._data[self._weight_index] *= scale
        return self

    ###########################################################################
    # SECTION: Constructors
    ###########################################################################
    @classmethod
    def _check_mom(cls, moments, mom_ndim, shape=None):
        """check moments for correct shape
        if moments is None, infer from shape[-mom_ndim:]
        if integer, convert to tuple
        """

        if moments is None:
            if shape is not None:
                if mom_ndim is None:
                    raise ValueError(
                        "must speficy either moments or shape and mom_ndim"
                    )
                moments = tuple(x - 1 for x in shape[-mom_ndim:])
            else:
                raise ValueError("must specify moments")

        if isinstance(moments, int):
            if mom_ndim is None:
                mom_ndim = 1
            moments = (moments,) * mom_ndim

        else:
            moments = tuple(moments)
            if mom_ndim is None:
                mom_ndim = len(moments)

        assert len(moments) == mom_ndim
        return moments

    @staticmethod
    def _datas_axis_to_first(datas, axis, mom_ndim):
        """move axis to first first position"""
        # NOTE: removinvg this. should be handles elsewhere
        # datas = np.asarray(datas)
        # ndim = datas.ndim - mom_ndim
        # if axis < 0:
        #     axis += ndim
        # assert 0 <= axis < ndim
        axis = normalize_axis_index(axis, datas.ndim - mom_ndim)
        if axis != 0:
            datas = np.moveaxis(datas, axis, 0)
        return datas, axis

    def _wrap_axis(self, axis, default=0, ndim=None):
        """wrap axis to positive value and check"""
        if axis is None:
            axis = default
        if ndim is None:
            ndim = self.val_ndim

        axis = normalize_axis_index(axis, ndim)
        # if axis < 0:
        #     axis += ndim
        # assert 0 <= axis < ndim
        return axis

    @classmethod
    def _mom_ndim_from_mom(cls, mom):
        if isinstance(mom, int):
            return 1
        elif isinstance(mom, tuple):
            return len(mom)
        else:
            raise ValueError("mom must be int or tuple")

    @classmethod
    def _choose_mom_ndim(cls, mom, mom_ndim):
        if mom is not None:
            mom_ndim = cls._mom_ndim_from_mom(mom)

        if mom_ndim is None:
            raise ValueError("must specify mom_ndim or mom")

        return mom_ndim

    @classmethod
    def from_data(
        cls,
        data,
        mom_ndim=None,
        mom=None,
        val_shape=None,
        copy=True,
        copy_kws=None,
        verify=True,
        dtype=None,
        **kws,
    ):
        """
        create new object with additional checks

        If pass `mom` and `shape`, make sure data conforms to this

        must pass either mom_ndim or mom
        """
        mom_ndim = cls._choose_mom_ndim(mom, mom_ndim)

        if verify:
            data = np.asarray(data, dtype=dtype)

        if val_shape is None:
            val_shape = data.shape[:-mom_ndim]
        mom = cls._check_mom(mom, mom_ndim, data.shape)

        if data.shape != val_shape + tuple(x + 1 for x in mom):
            raise ValueError(f"{data.shape} does not conform to {val_shape} and {mom}")

        if copy:
            if copy_kws is None:
                copy_kws = {}
            data = data.copy(**copy_kws)

        return cls(data=data, mom_ndim=mom_ndim, **kws)

    @classmethod
    def from_datas(
        cls,
        datas,
        mom_ndim=None,
        axis=0,
        mom=None,
        val_shape=None,
        dtype=None,
        verify=True,
        **kws,
    ):
        """
        Data should have val_shape

        [..., moments] (axis!= -1)

        [..., moment, axis] (axis == -1)
        """

        mom_ndim = cls._choose_mom_ndim(mom, mom_ndim)

        if verify:
            datas = np.asarray(datas, dtype=dtype)
        datas, axis = cls._datas_axis_to_first(datas, axis, mom_ndim)

        if val_shape is None:
            val_shape = datas.shape[1:-mom_ndim]

        mom = cls._check_mom(mom, mom_ndim, datas.shape)
        assert datas.shape[1:] == val_shape + tuple(x + 1 for x in mom)

        if dtype is None:
            dtype = datas.dtype

        return cls.zeros(
            shape=datas.shape[1:], mom_ndim=mom_ndim, dtype=dtype, **kws
        ).push_datas(datas=datas, axis=0)

    @classmethod
    def from_vals(
        cls,
        x,
        w=None,
        axis=0,
        mom=2,
        val_shape=None,
        dtype=None,
        broadcast=False,
        **kws,
    ):

        mom_ndim = cls._mom_ndim_from_mom(mom)
        x0 = x if mom_ndim == 1 else x[0]
        if val_shape is None:
            val_shape = _shape_reduce(x0.shape, axis)
        if dtype is None:
            dtype = x0.dtype

        return cls.zeros(val_shape=val_shape, mom=mom, dtype=dtype, **kws).push_vals(
            x=x, axis=axis, w=w, broadcast=broadcast
        )

    @classmethod
    def from_resample_vals(
        cls,
        x,
        freq=None,
        indices=None,
        nrep=None,
        w=None,
        axis=0,
        mom=2,
        dtype=None,
        broadcast=False,
        parallel=True,
        resample_kws=None,
        **kws,
    ):

        mom_ndim = cls._mom_ndim_from_mom(mom)

        x0 = x if mom_ndim == 1 else x[0]
        freq = randsamp_freq(
            nrep=nrep,
            freq=freq,
            indices=indices,
            size=x0.shape[axis],
            check=True,
        )

        if resample_kws is None:
            resample_kws = {}

        data = resample_vals(
            x,
            freq=freq,
            mom=mom,
            axis=axis,
            w=w,
            mom_ndim=mom_ndim,
            parallel=parallel,
            **resample_kws,
            broadcast=broadcast,
        )
        return cls.from_data(data, mom_ndim=mom_ndim, copy=False, **kws)

    @classmethod
    def from_raw(cls, raw, mom_ndim=None, mom=None, val_shape=None, dtype=None, **kws):
        """create object from raw

        must specify either `mom_ndim` or `mom`
        """

        mom_ndim = cls._choose_mom_ndim(mom, mom_ndim)

        if mom_ndim == 1:
            func = convert.to_central_moments
        elif mom_ndim == 2:
            func = convert.to_central_comoments
        data = func(raw)

        return cls.from_data(
            data,
            mom_ndim=mom_ndim,
            mom=mom,
            val_shape=val_shape,
            dtype=dtype,
            copy=False,
            **kws,
        )

    @classmethod
    def from_raws(
        cls, raws, mom_ndim=None, mom=None, axis=0, val_shape=None, dtype=None, **kws
    ):
        mom_ndim = cls._choose_mom_ndim(mom, mom_ndim)

        if mom_ndim == 1:
            func = convert.to_central_moments
        elif mom_ndim == 2:
            func = convert.to_central_comoments
        datas = func(raws)
        return cls.from_datas(
            datas=datas,
            axis=axis,
            mom_ndim=mom_ndim,
            mom=mom,
            val_shape=val_shape,
            dtype=dtype,
            **kws,
        )

    ###########################################################################
    # SECTION: Manipulation
    ###########################################################################
    @property
    def _is_vector(self):
        return self.val_ndim > 0

    def _raise_if_scalar(self, message=None):
        if not self._is_vector:
            if message is None:
                message = "not implemented for scalar"
            raise ValueError(message)

    # Universal reducers
    def resample_and_reduce(
        self,
        freq=None,
        indices=None,
        nrep=None,
        axis=None,
        parallel=True,
        resample_kws=None,
        **kws,
    ):
        """
        bootstrap resample and reduce

        Parameter
        ----------
        freq : array-like, shape=(nrep, nrec), optional
            frequence table.  freq[i, j] is the weight of the jth record to the ith
            replicate indices : array-like, shape=(nrep, nrec), optional
            resampling array.  idx[i, j] is the record index of the original array to
            place in new sample[i, j]. if specified, create freq array from idx
        nrep : int, optional
            if specified, create idx array with this number of replicates
        axis : int, Default=0
            axis to resample and reduce along
        parallel : bool, default=True
            flags to `numba.njit`
        resample_kws : dict
            extra arguments to `cmomy.resample.resample_and_reduce`
        args : tuple
            extra positional arguments to from_data method
        kwargs : dict
            extra key-word arguments to from_data method
        """
        self._raise_if_scalar()
        axis = self._wrap_axis(axis)
        if resample_kws is None:
            resample_kws = {}

        freq = randsamp_freq(
            nrep=nrep, indices=indices, freq=freq, size=self.val_shape[axis], check=True
        )
        data = resample_data(
            self.data, freq, mom=self.mom, axis=axis, parallel=parallel, **resample_kws
        )
        return type(self).from_data(data, mom_ndim=self.mom_ndim, copy=False, **kws)

    def resample(self, indices, axis=0, first=True, **kws):
        """
        create a new object sampled from index

        Parameters
        ----------
        indicies : array-like
        axis : int, default=0
            axis to resample
        first : bool, default=True
            if True, and axis != 0, the move the axis to first position.
            This makes results similar to resample and reduce
            If `first` False, then resampled array can have odd shape

        Returns
        -------
        output : accumulator object
        """
        self._raise_if_scalar()
        axis = self._wrap_axis(axis)

        data = self.data
        if first and axis != 0:
            data = np.moveaxis(data, axis, 0)
            axis = 0

        out = np.take(data, indices, axis=axis)
        return type(self)(data=out, mom_ndim=self.mom_ndim, **kws)

    def reduce(self, axis=0, **kws):
        """
        create new object reducealong axis
        """
        self._raise_if_scalar()
        axis = self._wrap_axis(axis)
        return type(self).from_datas(
            self.values, mom_ndim=self.mom_ndim, axis=axis, **kws
        )

    def block(self, block_size=None, axis=None, **kws):
        """
        block average reduction

        Parameters
        ----------
        block_size : int
            number of consecutive records to combine
        axis : int, default=0
            axis to reduce along
        kws : dict
            extral key word arguments to `from_datas` method
        """

        self._raise_if_scalar()

        axis = self._wrap_axis(axis)
        data = self.data

        # move axis to first
        if axis != 0:
            data = np.moveaxis(data, axis, 0)

        n = data.shape[0]

        if block_size is None:
            block_size = n
            nblock = 1

        else:
            nblock = n // block_size

        datas = data[: (nblock * block_size), ...].reshape(
            (nblock, block_size) + data.shape[1:]
        )
        return type(self).from_datas(datas=datas, mom_ndim=self.mom_ndim, axis=1, **kws)

    def reshape(self, shape, copy=True, copy_kws=None, **kws):
        """
        create a new object with reshaped data

        Parameters
        ---------
        shape : tuple
            shape of values part of data
        """
        self._raise_if_scalar()
        new_shape = shape + self.mom_shape
        data = self._data.reshape(new_shape)
        return self.new_like(
            data=data, verify=False, check=False, copy=copy, copy_kws=copy_kws, **kws
        )

    def moveaxis(self, source, destination, copy=True, copy_kws=None, **kws):
        """
        move axis from source to destination
        """
        self._raise_if_scalar()

        def _check_val(v):
            if isinstance(v, int):
                v = (v,)
            else:
                v = tuple(v)
            return tuple(self._wrap_axis(x) for x in v)

        source = _check_val(source)
        destination = _check_val(destination)
        data = np.moveaxis(self.data, source, destination)

        # use from data for extra checks
        # return self.new_like(data=data, copy=copy, *args, **kwargs)
        return type(self).from_data(
            data,
            mom=self.mom,
            mom_ndim=self.mom_ndim,
            val_shape=data.shape[: -self.mom_ndim],
            copy=copy,
            copy_kws=copy_kws,
            **kws,
        )

    # --------------------------------------------------
    # mom_ndim == 1 specific
    # --------------------------------------------------

    @staticmethod
    def _raise_if_not_1d(mom_ndim):
        if mom_ndim != 1:
            raise NotImplementedError("only available for mom_ndim == 1")

    # special, 1d only methods
    def push_stat(self, a, v=0.0, w=None, broadcast=True):
        self._raise_if_not_1d(self.mom_ndim)

        ar, target = self.check_val(a, target="val")
        vr = self.check_var(v, broadcast=broadcast)
        wr = self.check_weight(w, target=target)
        self._push.stat(self._data_flat, wr, ar, vr)
        return self

    def push_stats(self, a, v=0.0, w=None, axis=0, broadcast=True):
        self._raise_if_not_1d(self.mom_ndim)

        ar, target = self.check_vals(a, target="vals", axis=axis)
        vr = self.check_vars(v, target=target, axis=axis, broadcast=broadcast)
        wr = self.check_weights(w, target=target, axis=axis)
        self._push.stats(self._data_flat, wr, ar, vr)
        return self

    @classmethod
    def from_stat(cls, a, v=0.0, w=None, mom=2, val_shape=None, dtype=None, **kws):
        """
        object from single weight, average, variance/covariance
        """
        mom_ndim = cls._mom_ndim_from_mom(mom)
        cls._raise_if_not_1d(mom_ndim)

        if val_shape is None:
            val_shape = a.shape
        if dtype is None:
            dtype = a.dtype

        return cls.zeros(val_shape=val_shape, mom=mom, dtype=dtype, **kws).push_stat(
            w=w, a=a, v=v
        )

    @classmethod
    def from_stats(
        cls,
        a,
        v=0.0,
        w=None,
        axis=0,
        mom=2,
        val_shape=None,
        dtype=None,
        **kws,
    ):
        """
        object from several weights, averages, variances/covarainces along axis
        """

        mom_ndim = cls._mom_ndim_from_mom(mom)
        cls._raise_if_not_1d(mom_ndim)

        # get val_shape
        if val_shape is None:
            val_shape = _shape_reduce(a.shape, axis)
        return cls.zeros(val_shape=val_shape, dtype=dtype, mom=mom, **kws).push_stats(
            a=a, v=v, w=w, axis=axis
        )



"""
low level routines to do pushing
"""
from __future__ import absolute_import

from collections import namedtuple

from .options import OPTIONS
from .utils import factory_binomial, myjit

# from functools import lru_cache


# Maximum binomial factor
_bfac = factory_binomial(OPTIONS["nmax"])


###############################################################################
# Moments
###############################################################################


@myjit
def _push_val(data, w, x):

    if w == 0.0:
        return

    order = data.shape[0] - 1

    data[0] += w
    alpha = w / data[0]
    one_alpha = 1.0 - alpha

    delta = x - data[1]
    incr = delta * alpha

    data[1] += incr
    if order == 1:
        return

    for a in range(order, 2, -1):
        # c > 1
        tmp = 0.0
        delta_b = 1.0
        alpha_b = 1.0
        minus_b = 1.0
        one_alpha_b = 1.0
        for b in range(0, a - 1):
            c = a - b
            tmp += _bfac[a, b] * delta_b * (minus_b * alpha_b * one_alpha * data[c])
            delta_b *= delta
            alpha_b *= alpha
            one_alpha_b *= one_alpha
            minus_b *= -1.0

        # c == 0
        # b = a
        tmp += delta * alpha * one_alpha * delta_b * (-minus_b * alpha_b + one_alpha_b)

        data[a] = tmp

    data[2] = one_alpha * (data[2] + delta * incr)


@myjit
def _push_vals(data, W, X):
    ns = X.shape[0]
    for s in range(ns):
        _push_val(data, W[s], X[s])


@myjit
def _push_vals_scale(data, W, X, scale):
    ns = X.shape[0]
    for s in range(ns):
        f = scale[s]
        if f == 0:
            continue
        _push_val(data, W[s] * f, X[s])


@myjit
def _push_stat(data, w, a, v):
    """
    w : weight
    a : average
    v[i] : <dx**(i+2)>

    scale : parameter to rescale the weight
    """
    if w == 0:
        return

    order = data.shape[0] - 1

    data[0] += w

    alpha = w / data[0]
    one_alpha = 1.0 - alpha
    delta = a - data[1]
    incr = delta * alpha

    data[1] += incr

    if order == 1:
        return

    for a1 in range(order, 2, -1):
        tmp = 0.0
        delta_b = 1.0
        alpha_b = 1.0
        minus_b = 1.0
        one_alpha_b = 1.0
        for b in range(0, a1 - 1):
            c = a1 - b
            tmp += (
                _bfac[a1, b]
                * delta_b
                * (
                    minus_b * alpha_b * one_alpha * data[c]
                    + one_alpha_b * alpha * v[c - 2]
                )
            )
            delta_b *= delta
            alpha_b *= alpha
            one_alpha_b *= one_alpha
            minus_b *= -1.0

        c = 0
        b = a1 - c
        tmp += delta * alpha * one_alpha * delta_b * (-minus_b * alpha_b + one_alpha_b)
        data[a1] = tmp

    data[2] = v[0] * alpha + one_alpha * (data[2] + delta * incr)


@myjit
def _push_stats(data, W, A, V):
    ns = A.shape[0]
    for s in range(ns):
        _push_stat(data, W[s], A[s], V[s, ...])


@myjit
def _push_data(data, data_in):
    _push_stat(data, data_in[0], data_in[1], data_in[2:])


@myjit
def _push_datas(data, data_in):
    ns = data_in.shape[0]
    for s in range(ns):
        _push_stat(data, data_in[s, 0], data_in[s, 1], data_in[s, 2:])


@myjit
def _push_data_scale(data, data_in, scale):
    _push_stat(data, data_in[0] * scale, data_in[1], data_in[2:])


@myjit
def _push_datas_scale(data, data_in, scale):
    ns = data_in.shape[0]
    for s in range(ns):
        f = scale[s]
        if f == 0:
            continue
        _push_stat(data, data_in[s, 0] * f, data_in[s, 1], data_in[s, 2:])


# Vector
@myjit
def _push_val_vec(data, w, x):
    nv = data.shape[0]
    for k in range(nv):
        _push_val(data[k, :], w[k], x[k])


@myjit
def _push_vals_vec(data, W, X):
    ns = X.shape[0]
    nv = data.shape[0]
    for s in range(ns):
        for k in range(nv):
            _push_val(data[k, :], W[s, k], X[s, k])


@myjit
def _push_stat_vec(data, w, a, v):
    nv = data.shape[0]
    for k in range(nv):
        _push_stat(data[k, :], w[k], a[k], v[k, :])


@myjit
def _push_stats_vec(data, W, A, V):
    # V[sample, moment-2, value]
    ns = A.shape[0]
    nv = data.shape[0]
    for s in range(ns):
        for k in range(nv):
            _push_stat(data[k, :], W[s, k], A[s, k], V[s, k, :])


@myjit
def _push_data_vec(data, data_in):
    nv = data.shape[0]
    for k in range(nv):
        _push_data(data[k, :], data_in[k, :])


@myjit
def _push_datas_vec(data, Data_in):
    ns = Data_in.shape[0]
    nv = data.shape[0]
    for s in range(ns):
        for k in range(nv):
            _push_data(data[k, :], Data_in[s, k, :])


@myjit
def _push_vals_scale_vec(data, W, X, scale):
    ns = X.shape[0]
    nv = data.shape[0]
    for s in range(ns):
        f = scale[s]
        if f == 0:
            continue
        for k in range(nv):
            _push_val(data[k, :], W[s, k] * f, X[s, k])


@myjit
def _push_data_scale_vec(data, data_in, scale):
    nv = data.shape[0]
    if scale != 0:
        for k in range(nv):
            _push_data_scale(data[k, :], data_in[k, :], scale)


@myjit
def _push_datas_scale_vec(data, Data_in, scale):
    ns = Data_in.shape[0]
    nv = data.shape[0]
    for s in range(ns):
        f = scale[s]
        if f == 0:
            continue
        for k in range(nv):
            _push_data_scale(data[k, :], Data_in[s, k, :], f)


######################################################################
# Covariance "stuff"
######################################################################


@myjit
def _push_val_cov(data, w, x0, x1):

    if w == 0.0:
        return

    order0 = data.shape[0] - 1
    order1 = data.shape[1] - 1

    data[0, 0] += w
    alpha = w / data[0, 0]
    one_alpha = 1.0 - alpha

    delta0 = x0 - data[1, 0]
    delta1 = x1 - data[0, 1]

    incr0 = delta0 * alpha
    incr1 = delta1 * alpha

    # NOTE: decided to force order > 1
    # otherwise, this is just normal variance
    # if order0 > 0:
    #     data[1, 0] += incr0
    # if order1 > 0:
    #     data[0, 1] += incr1
    data[1, 0] += incr0
    data[0, 1] += incr1

    a0_min = max(0, 2 - order1)
    for a0 in range(order0, a0_min - 1, -1):
        a1_min = max(0, 2 - a0)
        for a1 in range(order1, a1_min - 1, -1):
            tmp = 0.0
            delta0_b0 = 1.0
            alpha_b0 = 1.0
            minus_b0 = 1.0
            one_alpha_b0 = 1.0
            for b0 in range(0, a0 + 1):
                c0 = a0 - b0
                f0 = _bfac[a0, b0]

                delta1_b1 = 1.0
                alpha_bb = alpha_b0
                minus_bb = minus_b0
                one_alpha_bb = one_alpha_b0
                for b1 in range(0, a1 + 1):
                    c1 = a1 - b1
                    cs = c0 + c1
                    if cs == 0:
                        tmp += (
                            delta0_b0
                            * delta1_b1
                            * (minus_bb * alpha_bb * one_alpha + one_alpha_bb * alpha)
                        )
                    elif cs != 1:
                        tmp += (
                            f0
                            * _bfac[a1, b1]
                            * delta0_b0
                            * delta1_b1
                            * (minus_bb * alpha_bb * one_alpha * data[c0, c1])
                        )
                    delta1_b1 *= delta1
                    alpha_bb *= alpha
                    one_alpha_bb *= one_alpha
                    minus_bb *= -1

                delta0_b0 *= delta0
                alpha_b0 *= alpha
                minus_b0 *= -1
                one_alpha_b0 *= one_alpha

            data[a0, a1] = tmp


@myjit
def _push_vals_cov(data, W, X1, X2):
    ns = X1.shape[0]
    for s in range(ns):
        _push_val_cov(data, W[s], X1[s], X2[s])


@myjit
def _push_vals_scale_cov(data, W, X1, X2, scale):
    ns = X1.shape[0]
    for s in range(ns):
        f = scale[s]
        if f == 0:
            continue
        _push_val_cov(data, W[s] * f, X1[s], X2[s])


@myjit
def _push_data_scale_cov(data, data_in, scale):

    w = data_in[0, 0] * scale
    if w == 0.0:
        return

    order0 = data.shape[0] - 1
    order1 = data.shape[1] - 1

    data[0, 0] += w
    alpha = w / data[0, 0]
    one_alpha = 1.0 - alpha

    delta0 = data_in[1, 0] - data[1, 0]
    delta1 = data_in[0, 1] - data[0, 1]

    incr0 = delta0 * alpha
    incr1 = delta1 * alpha

    # NOTE : decided to force all orders >0
    # if order0 > 0:
    #     data[1, 0] += incr0
    # if order1 > 0:
    #     data[0, 1] += incr1
    data[1, 0] += incr0
    data[0, 1] += incr1

    a0_min = max(0, 2 - order1)
    for a0 in range(order0, a0_min - 1, -1):
        a1_min = max(0, 2 - a0)
        for a1 in range(order1, a1_min - 1, -1):
            # Alternative
            tmp = 0.0
            delta0_b0 = 1.0
            alpha_b0 = 1.0
            minus_b0 = 1.0
            one_alpha_b0 = 1.0
            for b0 in range(0, a0 + 1):
                c0 = a0 - b0
                f0 = _bfac[a0, b0]

                delta1_b1 = 1.0
                alpha_bb = alpha_b0
                minus_bb = minus_b0
                one_alpha_bb = one_alpha_b0
                for b1 in range(0, a1 + 1):
                    c1 = a1 - b1
                    cs = c0 + c1
                    if cs == 0:
                        tmp += (
                            delta0_b0
                            * delta1_b1
                            * (minus_bb * alpha_bb * one_alpha + one_alpha_bb * alpha)
                        )
                    elif cs != 1:
                        tmp += (
                            f0
                            * _bfac[a1, b1]
                            * delta0_b0
                            * delta1_b1
                            * (
                                minus_bb * alpha_bb * one_alpha * data[c0, c1]
                                + one_alpha_bb * alpha * data_in[c0, c1]
                            )
                        )
                    delta1_b1 *= delta1
                    alpha_bb *= alpha
                    one_alpha_bb *= one_alpha
                    minus_bb *= -1

                delta0_b0 *= delta0
                alpha_b0 *= alpha
                minus_b0 *= -1
                one_alpha_b0 *= one_alpha

            data[a0, a1] = tmp


@myjit
def _push_data_cov(data, data_in):
    _push_data_scale_cov(data, data_in, 1.0)


@myjit
def _push_datas_cov(data, datas):
    ns = datas.shape[0]
    for s in range(ns):
        _push_data_scale_cov(data, datas[s], 1.0)


@myjit
def _push_datas_scale_cov(data, datas, scale):
    ns = datas.shape[0]
    for s in range(ns):
        f = scale[s]
        if f == 0:
            continue
        _push_data_scale_cov(data, datas[s], f)


# Vector
@myjit
def _push_val_cov_vec(data, w, x0, x1):
    nv = data.shape[0]
    for k in range(nv):
        _push_val_cov(data[k, ...], w[k], x0[k], x1[k])


@myjit
def _push_vals_cov_vec(data, W, X0, X1):
    nv = data.shape[0]
    ns = X0.shape[0]
    for s in range(ns):
        for k in range(nv):
            _push_val_cov(data[k, ...], W[s, k], X0[s, k], X1[s, k])


@myjit
def _push_vals_scale_cov_vec(data, W, X0, X1, scale):
    nv = data.shape[0]
    ns = X0.shape[0]
    for s in range(ns):
        f = scale[s]
        if f == 0:
            continue
        for k in range(nv):
            _push_val_cov(data[k, ...], W[s, k] * f, X0[s, k], X1[s, k])


@myjit
def _push_data_cov_vec(data, data_in):
    nv = data.shape[0]
    for k in range(nv):
        _push_data_scale_cov(data[k, ...], data_in[k, ...], 1.0)


@myjit
def _push_datas_cov_vec(data, Datas):
    nv = data.shape[0]
    ns = Datas.shape[0]
    for s in range(ns):
        for k in range(nv):
            _push_data_scale_cov(data[k, ...], Datas[s, k, ...], 1.0)


@myjit
def _push_data_scale_cov_vec(data, data_in, scale):
    nv = data.shape[0]
    if scale > 0:
        for k in range(nv):
            _push_data_scale_cov(data[k, ...], data_in[k, ...], scale)


@myjit
def _push_datas_scale_cov_vec(data, Datas, scale):
    nv = data.shape[0]
    ns = Datas.shape[0]
    for s in range(ns):
        f = scale[s]
        if f == 0:
            continue
        for k in range(nv):
            _push_data_scale_cov(data[k, ...], Datas[s, k, ...], f)


# named tuple for pushers


Pusher = namedtuple("Pusher", ["val", "vals", "stat", "stats", "data", "datas"])

pusher_scalar = Pusher(
    val=_push_val,
    vals=_push_vals,
    stat=_push_stat,
    stats=_push_stats,
    data=_push_data,
    datas=_push_datas,
)

pusher_vector = Pusher(
    val=_push_val_vec,
    vals=_push_vals_vec,
    stat=_push_stat_vec,
    stats=_push_stats_vec,
    data=_push_data_vec,
    datas=_push_datas_vec,
)


pusher_cov_scalar = Pusher(
    val=_push_val_cov,
    vals=_push_vals_cov,
    stat=None,
    stats=None,
    data=_push_data_cov,
    datas=_push_datas_cov,
)

pusher_cov_vector = Pusher(
    val=_push_val_cov_vec,
    vals=_push_vals_cov_vec,
    stat=None,
    stats=None,
    data=_push_data_cov_vec,
    datas=_push_datas_cov_vec,
)


def factory_pushers(cov=False, vec=False):
    if cov:
        if vec:
            return pusher_cov_vector
        else:
            return pusher_cov_scalar
    else:
        if vec:
            return pusher_vector
        else:
            return pusher_scalar


def factory_pusher_datas_scale(cov=False, vec=False):
    if cov:
        if vec:
            return _push_datas_scale_cov_vec
        else:
            return _push_datas_scale_cov

    else:
        if vec:
            return _push_datas_scale_vec
        else:
            return _push_datas_scale


def factory_pusher_vals_scale(cov=False, vec=False):
    if cov:
        if vec:
            return _push_vals_scale_cov_vec
        else:
            return _push_vals_scale_cov

    else:
        if vec:
            return _push_vals_scale_vec
        else:
            return _push_vals_scale



"""
Routines to perform resampling
"""
from __future__ import absolute_import

import numpy as np
import xarray as xr

from ._resample import factory_resample_data, factory_resample_vals
from .utils import _axis_expand_broadcast, myjit

###############################################################################
# resampling
###############################################################################


@myjit
def _randsamp_freq_out(freq):

    nrep = freq.shape[0]
    ndat = freq.shape[1]
    for i in range(nrep):
        for j in range(ndat):
            index = np.random.randint(0, ndat)
            freq[i, index] += 1


@myjit
def _randsamp_freq_indices(indices, freq):
    assert freq.shape == indices.shape
    nrep, ndat = freq.shape
    for r in range(nrep):
        for d in range(ndat):
            idx = indices[r, d]
            freq[r, idx] += 1


def randsamp_freq(
    nrep=None, size=None, indices=None, transpose=False, freq=None, check=False
):
    """
    produce a random sample for bootstrapping

    Parameters
    ----------
    size : int, optional
        data dimension size
    freq : array-like, shape=(nrep, size), optional
        if passed, use this frequency array.
        overides size
    indices : array-like, shape=(nrep, size), optional
        if passed and `freq` is `None`, construct frequency
        array from this indices array

    nrep : int, optional
        if `freq` and `indices` are `None`, construct
        sample with this number of repititions
    indices : array-like, optional
        if passed, build frequency table based on this sampling.
        shape = (nrep, ndat)
    freq : array-like, optional
        if passed, use this frequency array
    transpose : bool
        see output
    check : bool, default=False
        if `check` is `True`, then check `freq` and `indices` against `size` and `nrep`

    Returns
    -------
    output : frequency table
        if not transpose: output.shape == (nrep, size)
        if tranpose, output.shae = (size, nrep)

    """

    def _array_check(x, name=""):
        x = np.asarray(x, dtype=np.int64)
        if check:
            if nrep is not None:
                if x.shape[0] != nrep:
                    raise ValueError("{} has wrong nrep".format(name))

            assert size is not None
            if x.shape[1] != size:
                raise ValueError("{} has wrong size".format(name))
        return x

    if freq is not None:
        freq = _array_check(freq, "freq")

    elif indices is not None:
        indices = _array_check(indices, "indices")
        freq = np.zeros(indices.shape, dtype=np.int64)
        _randsamp_freq_indices(indices, freq)

    elif nrep is not None and size is not None:
        freq = np.zeros((nrep, size), dtype=np.int64)
        _randsamp_freq_out(freq)

    else:
        raise ValueError("must specify freq, indices, or nrep and size")

    if transpose:
        freq = freq.T
    return freq


def resample_data(
    data,
    freq,
    mom,
    axis=0,
    dtype=None,
    order=None,
    parallel=True,
    out=None,
):
    """
    resample data according to frequency table

    Parameters
    ----------
    data : array-like
        central mom array to be resampled
    freq : array-like
        frequency array with shape (nrep, data.shape[axis])
    mom : int or array-like
        if int or length 1, then data contains central mom.
        if length is 2, then data contains central comom
    axis : int, default=0
        axis to reduce along
    parallel : bool
        options for jitting pusher
    dtype, order : options to np.asarray
    out : optional output

    Returns
    -------
    output : array
        output shape is `(nrep,) + shape + mom`, where shape is
        the shape of data less axis, and mom is the shape of the resulting mom.
    """

    if isinstance(mom, int):
        mom = (mom,)

    # check inputs
    data = np.asarray(data, dtype=dtype, order=order)
    freq = np.asarray(freq, dtype=np.int64, order=order)

    if dtype is None:
        dtype = data.dtype

    nrep, ndat = freq.shape
    ndim = data.ndim - len(mom)
    if axis < 0:
        axis += ndim
    assert 0 <= axis < ndim

    if axis != 0:
        data = np.moveaxis(data, axis, 0)

    shape = data.shape[1 : -len(mom)]
    mom_shape = tuple(x + 1 for x in mom)

    assert data.shape == (ndat,) + shape + mom_shape

    # output
    out_shape = (nrep,) + data.shape[1:]
    if out is None:
        out = np.empty(out_shape, dtype=dtype)
    assert out.shape == out_shape

    # resahpe
    if shape == ():
        meta_reshape = ()
    else:
        meta_reshape = (np.prod(shape),)

    data_reshape = (ndat,) + meta_reshape + mom_shape
    out_reshape = (nrep,) + meta_reshape + mom_shape

    datar = data.reshape(data_reshape)
    outr = out.reshape(out_reshape)

    resample = factory_resample_data(
        cov=len(mom) > 1, vec=len(shape) > 0, parallel=parallel
    )

    outr.fill(0.0)
    resample(datar, freq, outr)

    return out


def resample_vals(
    x,
    freq,
    mom,
    axis=0,
    w=None,
    mom_ndim=None,
    broadcast=False,
    dtype=None,
    order=None,
    parallel=True,
    out=None,
):
    """
    resample data according to frequency table
    """

    if isinstance(mom, int):
        mom = (mom,) * 1
    assert isinstance(mom, tuple)

    if mom_ndim is None:
        mom_ndim = len(mom)
    assert len(mom) == mom_ndim
    mom_shape = tuple(x + 1 for x in mom)

    if mom_ndim == 1:
        y = None
    elif mom_ndim == 2:
        x, y = x
    else:
        raise ValueError("only mom_ndim <= 2 supported")

    cov = y is not None

    # check input data
    freq = np.asarray(freq, dtype=np.int64)
    nrep, ndat = freq.shape

    x = np.asarray(x, dtype=dtype, order=order)
    if dtype is None:
        dtype = x.dtype
    if w is None:
        w = np.ones_like(x)
    else:
        w = _axis_expand_broadcast(
            w, x.shape, axis, roll=False, dtype=dtype, order=order
        )

    if cov:
        y = _axis_expand_broadcast(
            y, x.shape, axis, roll=False, broadcast=broadcast, dtype=dtype, order=order
        )

    if axis != 0:
        x = np.moveaxis(x, axis, 0)
        w = np.moveaxis(w, axis, 0)
        if cov:
            y = np.moveaxis(y, axis, 0)

    assert len(x) == ndat

    # output
    shape = x.shape[1:]
    out_shape = (nrep,) + shape + mom_shape
    if out is None:
        out = np.empty(out_shape, dtype=dtype)
    assert out.shape == out_shape

    # reshape
    if shape == ():
        meta_reshape = ()
    else:
        meta_reshape = (np.prod(shape),)
    data_reshape = (ndat,) + meta_reshape
    out_reshape = (nrep,) + meta_reshape + mom_shape

    xr = x.reshape(data_reshape)
    wr = w.reshape(data_reshape)
    outr = out.reshape(out_reshape)
    if cov:
        yr = y.reshape(data_reshape)

    resample = factory_resample_vals(cov=cov, vec=len(shape) > 0, parallel=parallel)
    outr.fill(0.0)
    if cov:
        resample(wr, xr, yr, freq, outr)
    else:
        resample(wr, xr, freq, outr)

    return out


def bootstrap_confidence_interval(
    distribution, stats_val="mean", axis=0, alpha=0.05, style=None, **kws
):
    """
    Calculate the error bounds

    Parameters
    ----------
    distribution : array-like
        distribution of values to consider
    stats_val : array-like, {None, 'mean','median'}
        * None: percentiles, with value as median
        * array: perform pivotal error bounds (correct) with this as `value`.
        * mean: pivotal error bounds with mean as value
        * median: pivotal error bounds with median as value
    axis : int, default=0
        axis to analyze along
    alpha : float
        alpha value for confidence interval.
        Percent confidence = `100 * (1 - alpha)`
    kws : dict
        extra arguments to `numpy.percentile`
    style : {None, 'delta', 'pm'}
        controls style of output

    Returns
    -------
    out : array
        fist dimension will be statistics.  Other dimensions
        have shape of input less axis reduced over.
        Depending on `style` first dimension will be
        (note val is either stats_val or median):

    * None: [val, low, high]
    * delta:  [val, val-low, high - val]
    * pm : [val, (high - low) / 2]
    """

    if stats_val is None:
        p_low = 100 * (alpha / 2.0)
        p_mid = 50
        p_high = 100 - p_low
        val, low, high = np.percentile(
            a=distribution, q=[p_mid, p_low, p_high], axis=axis, **kws
        )

    else:
        if isinstance(stats_val, str):
            if stats_val == "mean":
                stats_val = np.mean(distribution, axis=axis)
            elif stats_val == "median":
                stats_val = np.median(distribution, axis=axis)
            else:
                raise ValueError("stats val should be None, mean, median, or an array")

        q_high = 100 * (alpha / 2.0)
        q_low = 100 - q_high
        val = stats_val
        low = 2 * stats_val - np.percentile(a=distribution, q=q_low, axis=axis, **kws)
        high = 2 * stats_val - np.percentile(a=distribution, q=q_high, axis=axis, **kws)

    if style is None:
        out = np.array([val, low, high])
    elif style == "delta":
        out = np.array([val, val - low, high - val])
    elif style == "pm":
        out = np.array([val, (high - low) / 2.0])
    return out


def xbootstrap_confidence_interval(
    x,
    stats_val="mean",
    axis=0,
    dim=None,
    alpha=0.05,
    style=None,
    bootstrap_dim="bootstrap",
    bootstrap_coords=None,
    **kws
):
    """
    xarray version of bootstrap_confidence_interval

    Parameters
    ----------
    dim : str
        if passed, use reduce along this dimension
    bootstrap_dim : str, default='bootstrap'
        name of new dimension.  If `bootstrap_dim` conflicts, then
        `new_name = dim + new_name`
    bootstrap_coords : array-like or str
        coords of new dimension.
        If `None`, use default names
        If string, use this for the 'values' name
    """

    if dim is not None:
        axis = x.get_axis_num(dim)
    else:
        dim = x.dims[axis]

    template = x.isel(**{dim: 0})

    if bootstrap_dim is None:
        bootstrap_dim = "bootstrap"

    if bootstrap_dim in template.dims:
        bootstrap_dim = "{}_{}".format(dim, bootstrap_dim)
    dims = (bootstrap_dim,) + template.dims

    if bootstrap_coords is None:
        if isinstance(stats_val, str):
            bootstrap_coords = stats_val
        else:
            bootstrap_coords = "stats_val"

    if isinstance(bootstrap_coords, str):
        if style is None:
            bootstrap_coords = [bootstrap_coords, "low", "high"]
        elif style == "delta":
            bootstrap_coords = [bootstrap_coords, "err_low", "err_high"]
        elif style == "pm":
            bootstrap_coords = [bootstrap_coords, "err"]

    if stats_val is not None and not isinstance(stats_val, str):
        stats_val = np.array(stats_val)

    out = bootstrap_confidence_interval(
        x, stats_val=stats_val, axis=axis, alpha=alpha, style=style, **kws
    )

    out = xr.DataArray(
        out,
        dims=dims,
        coords=template.coords,
        attrs=template.attrs,
        name=template.name,
        indexes=template.indexes,
    )
    if bootstrap_coords is not None:
        out.coords[bootstrap_dim] = bootstrap_coords
    return out



from setuptools import setup  # , find_packages

import versioneer

# with open("README.md", "r") as f:
#     long_description = f.read()

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    # name="cmomy",
    # version="0.0.2",
    # description="Central (co)moment calculation",
    # long_description=long_description,
    # long_description_content_type="text/markdown",
    # classifiers=[
    #     "Development Status :: 3 - Alpha",
    #     "Programming Language :: Python :: 3.6",
    #     "Programming Language :: Python :: 3.7",
    #     "Programming Language :: Python :: 3.8",
    # ],
    # keywords="data analysis",
    # url="https://github.com/wpk-nist-gov/cmomy",
    # author="William Krekelberg",
    # author_email="wpk@nist.gov",
    # license="NIST license https://www.nist.gov/director/licensing",
    # packages=find_packages(),
    # install_requires=["numpy", "numba", "xarray"],
    # testing
    # tests_require=["pytest"],
    # test_suite="pytest",
    # include_package_data=True,
    # zip_safe=True,
)



"""
Thin wrapper around central routines with xarray support
"""
from __future__ import absolute_import

import numpy as np
import xarray as xr

from . import central
from .cached_decorators import gcached
from .utils import _xr_order_like, _xr_wrap_like


###############################################################################
# central mom/comom routine
###############################################################################
def _xcentral_moments(
    x,
    mom,
    w=None,
    axis=0,
    last=True,
    mom_dims=None,
):

    assert isinstance(x, xr.DataArray)

    if isinstance(mom, tuple):
        mom = mom[0]

    if mom_dims is None:
        mom_dims = ("mom_0",)
    if isinstance(mom_dims, str):
        mom_dims = (mom_dims,)
    assert len(mom_dims) == 1

    if w is None:
        w = xr.ones_like(x)
    else:
        w = xr.DataArray(w).broadcast_like(x)

    if isinstance(axis, int):
        dim = x.dims[axis]
    else:
        dim = axis

    wsum = w.sum(dim=dim)
    wsum_inv = 1.0 / wsum

    xave = xr.dot(w, x, dims=dim) * wsum_inv

    p = xr.DataArray(np.arange(0, mom + 1), dims=mom_dims)
    dx = (x - xave) ** p
    out = xr.dot(w, dx, dims=dim) * wsum_inv

    out.loc[{mom_dims[0]: 0}] = wsum
    out.loc[{mom_dims[0]: 1}] = xave

    if last:
        out = out.transpose(..., *mom_dims)
    return out


def _xcentral_comoments(
    x,
    mom,
    w=None,
    axis=0,
    last=True,
    broadcast=False,
    mom_dims=None,
):
    """
    calculate central co-mom (covariance, etc) along axis
    """

    if isinstance(mom, int):
        mom = (mom,) * 2

    mom = tuple(mom)
    assert len(mom) == 2

    assert isinstance(x, tuple) and len(x) == 2
    x, y = x

    assert isinstance(x, xr.DataArray)

    if w is None:
        w = xr.ones_like(x)
    else:
        w = xr.DataArray(w).broadcast_like(x)

    if broadcast:
        y = xr.DataArray(y).broadcast_like(x)
    else:
        assert isinstance(y, xr.DataArray)

        y = y.transpose(*x.dims)
        assert y.shape == x.shape
        assert x.dims == y.dims

    if isinstance(axis, int):
        dim = x.dims[axis]
    else:
        dim = axis

    if mom_dims is None:
        mom_dims = ["mom_0", "mom_1"]
    assert len(mom_dims) == 2

    wsum = w.sum(dim=dim)
    wsum_inv = 1.0 / wsum

    x = (x, y)

    xave = [xr.dot(w, xx, dims=dim) * wsum_inv for xx in x]

    p = [
        xr.DataArray(np.arange(0, mom + 1), dims=dim) for mom, dim in zip(mom, mom_dims)
    ]

    dx = [(xx - xxave) ** pp for xx, xxave, pp in zip(x, xave, p)]

    out = xr.dot(w, dx[0], dx[1], dims=dim) * wsum_inv

    out.loc[{mom_dims[0]: 0, mom_dims[1]: 0}] = wsum
    out.loc[{mom_dims[0]: 1, mom_dims[1]: 0}] = xave[0]
    out.loc[{mom_dims[0]: 0, mom_dims[1]: 1}] = xave[1]

    if last:
        out = out.transpose(..., *mom_dims)
    return out


def xcentral_moments(
    x,
    mom,
    w=None,
    axis=0,
    last=True,
    mom_dims=None,
    broadcast=False,
):
    """
    calculate central mom along axis

    Parameters
    ----------
    x : xarray.DataArray or tuple of xarray.Datarray
        input data
    mom : int
        number of mom to calculate
    w : array-like, optional
        if passed, should be able to broadcast to `x`. An exception is if
        w is a 1d array with len(w) == x.shape[axis]. In this case,
        w will be reshaped and broadcast against x
    axis : int, default=0
        axis to reduce along
    last : bool, default=True
        if True, put mom as last dimension.
        Otherwise, mom will be in first dimension
    dtype, order : options to np.asarray
    out : array
        if present, use this for output data
        Needs to have shape of either (mom,) + shape or shape + (mom,)
        where shape is the shape of x with axis removed

    Returns
    -------
    output : array
        array of shape shape + (mom,) or (mom,) + shape depending on
        value of `last`, where `shape` is the shape of `x` with axis removed,
        i.e., shape=x.shape[:axis] + x.shape[axis+1:]. Assuming `last is True`,
        output[...,0] is the total weight (or count), output[...,1] is the mean
        value of x, output[...,n] with n>1 is the nth central moment
    """

    if isinstance(mom, int):
        mom = (mom,)

    kws = dict(x=x, mom=mom, w=w, axis=axis, last=last, mom_dims=mom_dims)
    if len(mom) == 1:
        func = _xcentral_moments
    else:
        func = _xcentral_comoments
        kws["broadcast"] = broadcast
    return func(**kws)


def _attributes_from_xr(da, dim=None, mom_dims=None, **kws):
    if isinstance(da, xr.DataArray):
        if dim is not None:
            # reduce along this dim
            da = da.isel(**{dim: 0}, drop=True)
        out = {k: getattr(da, k) if v is None else v for k, v in kws.items()}
    else:
        out = kws.copy()

    out["mom_dims"] = mom_dims
    return out


def _check_xr_input(x, axis=None, mom_dims=None, **kws):
    if isinstance(x, xr.DataArray):
        if axis is None:
            dim = None
        elif isinstance(axis, str):
            dim = axis
            axis = x.get_axis_num(dim)
        else:
            dim = x.dims[axis]

        values = x.values
    else:
        if axis is None:
            dim = None
        else:
            dim = axis
        values = x

    kws = _attributes_from_xr(x, dim=dim, mom_dims=mom_dims, **kws)
    return kws, axis, values


class xCentralMoments(central.CentralMoments):
    """
    xarray.DataArray wrapper of `cmomy.central.CentralMoments`

    Most methods are wrapped to accept `xarray.DataArray` object

    Notes:
    ------
    unlike xarray, most methods take only the `axis` parameter
    instead of both an `axis` (for positional) and `dim` (for names)
    parameter for reduction.  If `axis` is a integer, then positional
    reduction is assumed, otherwise named reduction is done.  In the
    case that `dims` have integer values, this will lead to only positional
    reduction.
    """

    __slots__ = "_xdata"

    def __init__(
        self,
        data,
        mom_ndim,
        dims=None,
        coords=None,
        attrs=None,
        name=None,
        indexes=None,
        mom_dims=None,
        **kws,
    ):

        if isinstance(data, xr.DataArray):
            if mom_dims is None:
                mom_dims = data.dims[-mom_ndim:]
            else:
                if isinstance(mom_dims, str):
                    mom_dims = [mom_dims]
                assert len(mom_dims) == mom_ndim
                order = (...,) + tuple(mom_dims)

                if data.dims != order:
                    data = data.transpose(*order)
            self._xdata = data

        else:
            data = self._wrap_data_with_xr(
                data=data,
                mom_ndim=mom_ndim,
                dims=dims,
                coords=coords,
                attrs=attrs,
                name=name,
                indexes=indexes,
                mom_dims=mom_dims,
            )

        self._xdata = data
        # TODO: data.data or data.values?
        super(xCentralMoments, self).__init__(data=data.data, mom_ndim=mom_ndim)

    @property
    def values(self):
        return self._xdata

    # xarray attriburtes
    @property
    def attrs(self):
        return self._xdata.attrs

    @property
    def dims(self):
        return self._xdata.dims

    @property
    def coords(self):
        return self._xdata.coords

    @property
    def name(self):
        return self._xdata.name

    @property
    def indexes(self):
        return self._xdata.indexes

    @property
    def sizes(self):
        return self._xdata.sizes

    ###########################################################################
    # SECTION: top level creation/copy/new
    ###########################################################################
    @gcached()
    def _template_val(self):
        """template for values part of data"""
        return self._xdata[self._weight_index]

    # def _wrap_like_template(self, x):
    #     return _wrap_like(self._template_val, x)

    def _wrap_like(self, x):
        return _xr_wrap_like(self._xdata, x)

    @property
    def val_dims(self):
        """dim names of values"""
        return self.dims[: -self.mom_ndim]

    @property
    def mom_dims(self):
        return self.dims[-self.mom_ndim :]

    @property
    def _one_like_val(self):
        return xr.ones_like(self._template_val)

    @property
    def _zeros_like_val(self):
        return xr.zeros_like(self._template_val)

    def new_like(
        self, data=None, verify=False, check=False, copy=False, copy_kws=None, **kws
    ):
        """
        create new object like self, with new data


        Parameters
        ----------
        data : array-like, optional
            data for new object
        verify : bool, default=False
            if True, pass data through np.asarray
        check : bool, default=True
            if True, then check that data has same total shape as self
        copy : bool, default=False
            if True, perform copy of data
        *args, **kwargs : extra arguments
            arguments to data.copy
        """

        if data is None:
            data = xr.zeros_like(self._xdata)

        else:
            if verify:
                if not isinstance(data, xr.DataArray):
                    data = self._wrap_like(data)

            if check:
                assert data.shape == self.shape
                data.dims == self.dims

            if copy:
                if copy_kws is None:
                    copy_kws = {}
                data = data.copy(**copy_kws)
        return type(self)(data=data, mom_ndim=self.mom_ndim, **kws)

    @classmethod
    def _wrap_data_with_xr(
        cls,
        data,
        mom_ndim,
        dims=None,
        coords=None,
        attrs=None,
        name=None,
        indexes=None,
        mom_dims=None,
    ):
        """
        given an array x, wrap with appropriate stuff
        for an xarray
        """

        ndim = data.ndim
        if dims is not None:
            if isinstance(dims, str):
                dims = [dims]
        else:
            dims = [f"dim_{i}" for i in range(ndim - mom_ndim)]
        dims = tuple(dims)

        if len(dims) == ndim:
            dims_total = dims
        elif len(dims) == ndim - mom_ndim:
            if mom_dims is None:
                # default mom dims
                mom_dims = [f"mom_{i}" for i in range(mom_ndim)]
            elif isinstance(mom_dims, str):
                mom_dims = [mom_dims]

            assert len(mom_dims) == mom_ndim
            dims_total = dims + tuple(mom_dims)
        else:
            raise ValueError("bad dims {}, moment_dims {}".format(dims, mom_dims))

        # xarray object
        return xr.DataArray(
            data,
            dims=dims_total,
            coords=coords,
            attrs=attrs,
            name=name,
            indexes=indexes,
        )

    def zeros_like(self, *args, **kwargs):
        return self.new_like(data=xr.zeros_like(self._xdata, *args, **kwargs))

    @classmethod
    def zeros(
        cls,
        mom=None,
        val_shape=None,
        mom_ndim=None,
        shape=None,
        dtype=None,
        zeros_kws=None,
        dims=None,
        coords=None,
        attrs=None,
        name=None,
        indexes=None,
        mom_dims=None,
        **kws,
    ):
        """
        create a new base object

        Parameters
        ----------
        shape : tuple, optional
            if passed, create object with this total shape
        mom : int or tuple
            moments.  if integer, then moments will be (mom,)
        val_shape : tuple, optional
            shape of values, excluding moments.  For example, if considering the average
            of observations `x`, then val_shape = x.shape.
            if not passed, then assume shape = ()
        dtype : nunpy dtype, default=float
        kwargs : dict
            extra arguments to numpy.zeros

        Returns
        -------
        object : instance of class `cls`

        Notes
        -----
        the resulting total shape of data is shape + mom_shape
        """

        return super(xCentralMoments, cls).zeros(
            mom=mom,
            val_shape=val_shape,
            mom_ndim=mom_ndim,
            shape=shape,
            dtype=dtype,
            zeros_kws=zeros_kws,
            dims=dims,
            coords=coords,
            attrs=attrs,
            name=name,
            indexes=indexes,
            mom_dims=mom_dims,
            **kws,
        )

    ###########################################################################
    # xarray specific methods
    ###########################################################################
    def _wrap_xarray_method(self, _method, *args, **kwargs):
        xdata = getattr(self._xdata, _method)(*args, **kwargs)
        return self.new_like(data=xdata)

    def assign_coords(self, coords=None, **coords_kwargs):
        return self._wrap_xarray_method("assign_coords", coords=coords, **coords_kwargs)

    def assign_attrs(self, *args, **kwargs):
        return self._wrap_xarray_method("assign_attrs", *args, **kwargs)

    def rename(self, new_name_or_name_dict=None, **names):
        return self._wrap_xarray_method(
            "rename", new_name_or_name_dict=new_name_or_name_dict, **names
        )

    def _wrap_xarray_method_from_data(
        self,
        _method,
        _data_copy=False,
        _data_order=False,
        _data_kws=None,
        *args,
        **kwargs,
    ):

        xdata = getattr(self._xdata, _method)(*args, **kwargs)
        if _data_order:
            xdata = xdata.transpose(..., *self.mom_dims)
        if _data_kws is None:
            _data_kws = {}
        _data_kws.setdefault("copy", _data_copy)
        _data_kws.setdefault("copy", _data_copy)

        return type(self).from_data(data=xdata, mom_ndim=self.mom_ndim, **_data_kws)

    def stack(
        self,
        dimensions=None,
        _data_copy=False,
        _data_kws=None,
        _data_order=True,
        **dimensions_kwargs,
    ):
        return self._wrap_xarray_method_from_data(
            "stack",
            _data_copy=_data_copy,
            _data_kws=_data_kws,
            _data_order=_data_order,
            **dimensions_kwargs,
        )

    def sel(
        self,
        indexers=None,
        method=None,
        tolerance=None,
        drop=False,
        _data_kws=None,
        _data_copy=False,
        _data_order=False,
        **indexers_kws,
    ):
        return self._wrap_xarray_method_from_data(
            "sel",
            _data_copy=_data_copy,
            _data_kws=_data_kws,
            _data_order=_data_order,
            indexers=indexers,
            method=method,
            tolerance=tolerance,
            drop=drop,
            **indexers_kws,
        )

    def isel(
        self,
        indexers=None,
        drop=False,
        _data_kws=None,
        _data_copy=False,
        _data_order=False,
        **indexers_kws,
    ):
        return self._wrap_xarray_method_from_data(
            "isel",
            _data_copy=_data_copy,
            _data_kws=_data_kws,
            _data_order=_data_order,
            indexers=indexers,
            drop=drop,
            **indexers_kws,
        )

    ###########################################################################
    # Push/verify
    ###########################################################################
    def _xverify_value(
        self, x, target=None, dim=None, broadcast=False, expand=False, shape_flat=None
    ):

        if isinstance(target, str):

            if dim is not None:
                if isinstance(dim, int):
                    dim = x.dims[dim]

            if target == "val":
                target = self.val_dims
            elif target == "vals":
                target = (dim,) + self.val_dims
            elif target == "data":
                target = self.dims
            elif target == "datas":
                target = (dim,) + self.dims

        if isinstance(target, tuple):
            # no broadcast in this cast
            target_dims = target

            target_shape = tuple(
                x.sizes[k] if k == dim else self.sizes[k] for k in target_dims
            )

            # make sure in correct order
            x = x.transpose(*target_dims)
            target_output = x
            values = x.values

        else:
            target_dims = target.dims
            target_shape = target.shape

            target_output = None

            if dim is not None:
                if isinstance(dim, int):
                    # this is a bit awkward, but
                    # should work
                    # assume we already have target in correct order
                    dim = target_dims[0]

            if isinstance(x, xr.DataArray):
                if broadcast:
                    x = x.broadcast_like(target)
                else:
                    x = x.transpose(*target_dims)

                values = x.values
            else:
                # only things this can be is either a scalor or
                # array with same size as target
                x = np.asarray(x)
                if x.shape == target.shape:
                    values = x
                    # have x -> target to get correct recs
                    x = target

                elif x.ndim == 0 and broadcast and expand:
                    x = xr.DataArray(x).broadcast_like(target)
                    values = x.values

                elif (
                    x.ndim == 1 and len(x) == target.sizes[dim] and broadcast and expand
                ):
                    x = xr.DataArray(x, dims=dim).broadcast_like(target)
                    values = x.values

        # check shape
        assert values.shape == target_shape
        if dim is None:
            nrec = ()
        else:
            nrec = (x.sizes[dim],)

        if shape_flat is not None:
            values = values.reshape(nrec + shape_flat)

        if values.ndim == 0:
            values = values[()]

        if target_output is None:
            return values
        else:
            return values, target_output

    def _verify_value(
        self,
        x,
        target=None,
        axis=None,
        broadcast=False,
        expand=False,
        shape_flat=None,
    ):
        if isinstance(x, xr.DataArray) or isinstance(target, xr.DataArray):
            return self._xverify_value(
                x,
                target=target,
                dim=axis,
                broadcast=broadcast,
                expand=expand,
                shape_flat=shape_flat,
            )

        else:
            return super(xCentralMoments, self)._verify_value(
                x,
                target=target,
                axis=axis,
                broadcast=broadcast,
                expand=expand,
                shape_flat=shape_flat,
            )

    def _single_index_selector(
        self, val, dim_combined="variable", coords_combined=None, select=True
    ):
        idxs = self._single_index(val)[-self.mom_ndim :]
        if coords_combined is None:
            coords_combined = self.mom_dims

        selector = {
            dim: (idx if self._mom_ndim == 1 else xr.DataArray(idx, dims=dim_combined))
            for dim, idx in zip(self.mom_dims, idxs)
        }
        if select:
            out = self.values.isel(**selector)
            if self._mom_ndim > 1:
                out = out.assign_coords(**{dim_combined: list(coords_combined)})
            return out
        else:
            return selector

        return selector

    # have to do some special stuff for mean/var
    # due to xarray special indexing
    def mean(self, dim_combined="variable", coords_combined=None):
        return self._single_index_selector(
            val=1, dim_combined=dim_combined, coords_combined=coords_combined
        )

    def var(self, dim_combined="variable", coords_combined=None):
        return self._single_index_selector(
            val=2, dim_combined=dim_combined, coords_combined=coords_combined
        )

    def cmom(self):
        return self._wrap_like(super(xCentralMoments, self).cmom())

    def rmom(self):
        return self._wrap_like(super(xCentralMoments, self).rmom())

    def resample_and_reduce(
        self,
        freq=None,
        indices=None,
        nrep=None,
        axis=None,
        rep_dim="rep",
        parallel=True,
        resample_kws=None,
        **kwargs,
    ):
        """
        bootstrap resample and reduce


        Parameters
        ----------
        rep_dim : str, default='rep'
            dimension name for resampled
            if 'dims' is not passed in kwargs, then reset dims
            with replicate dimension having name 'rep_dim',
            and all other dimensions have the same names as
            the parent object
        """

        if "dims" in kwargs:
            # passed explicit dims, so just use these.
            # NOTE, don't do anything special in this case
            pass
        else:
            axis = self._wrap_axis(axis)
            dims = list(self.values.dims)
            kwargs["dims"] = [rep_dim] + dims[:axis] + dims[axis + 1 :]

        return super(xCentralMoments, self).resample_and_reduce(
            freq=freq,
            indices=indices,
            nrep=nrep,
            axis=axis,
            parallel=parallel,
            resample_kws=resample_kws,
            **kwargs,
        )

    def block(self, block_size, axis=None, *args, **kwargs):
        """

        block average along an axis

        Parameters
        ----------
        block_size : int
            size of blocks to average over
        axis : str or int, default=0
            axis/dimension to block average over
        args : tuple
            positional arguments to CentralMomentsBase.block
        kwargs : dict
            key-word arguments to CentralMomentsBase.block
        """

        self._raise_if_scalar()
        axis = self._wrap_axis(axis)
        dim = self.dims[axis]

        template = self.values.isel(**{dim: 0})

        kwargs["dims"] = (dim,) + template.dims

        # properties after creation
        for k in ["coords", "attrs", "name", "indexes"]:
            kwargs[k] = getattr(template, k)

        return super(xCentralMoments, self).block(
            block_size=block_size, axis=axis, **kwargs
        )

    # def resample(self, indices, axis=0, first=True, **kws):
    #     self._raise_if_scalar()
    #     axis = self._wrap_axis(axis)
    #     dim = self.

    #     if not isinstance(indices, xr.DataArray):

    def _wrap_axis(self, axis, default=0, ndim=None):
        # if isinstance(axis, str):
        #     axis = self._xdata.get_axis_num(axis)
        # return super(xCentralMoments, self)._wrap_axis(
        #     axis=axis, default=default, ndim=ndim
        # )
        if isinstance(axis, str):
            return self._xdata.get_axis_num(axis)
        else:
            return super(xCentralMoments, self)._wrap_axis(
                axis=axis, default=default, ndim=ndim
            )

    @classmethod
    def from_data(
        cls,
        data,
        mom=None,
        mom_ndim=None,
        val_shape=None,
        dtype=None,
        copy=True,
        dims=None,
        attrs=None,
        coords=None,
        indexes=None,
        name=None,
        mom_dims=None,
    ):
        """
        object from data array


        Parameters
        ----------
        dims : tuple, optional
            dimension names for resulting object
        """

        kws = _attributes_from_xr(
            data,
            mom_dims=mom_dims,
            dims=dims,
            attrs=attrs,
            coords=coords,
            indexes=indexes,
            name=name,
        )

        verify = not isinstance(data, xr.DataArray)
        return super(xCentralMoments, cls).from_data(
            data=data,
            mom=mom,
            mom_ndim=mom_ndim,
            val_shape=val_shape,
            dtype=dtype,
            copy=copy,
            verify=verify,
            **kws,
        )

    @classmethod
    def from_datas(
        cls,
        datas,
        mom=None,
        mom_ndim=None,
        axis=0,
        val_shape=None,
        dtype=None,
        dims=None,
        attrs=None,
        coords=None,
        indexes=None,
        name=None,
        mom_dims=None,
    ):
        """
        Parameters
        ----------
        dims : tuple, optional
            dimension names.
            Note that this does not include the dimension reduced over.
        """

        kws, axis, values = _check_xr_input(
            datas,
            axis=axis,
            mom_dims=mom_dims,
            dims=dims,
            attrs=attrs,
            coords=coords,
            indexes=indexes,
            name=name,
        )

        return super(xCentralMoments, cls).from_datas(
            datas=values,
            mom=mom,
            mom_ndim=mom_ndim,
            axis=axis,
            val_shape=val_shape,
            dtype=dtype,
            **kws,
        )

    def to_raw(self):
        return self._wrap_like(super(xCentralMoments, self).to_raw())

    @classmethod
    def from_raw(
        cls,
        raw,
        mom=None,
        mom_ndim=None,
        val_shape=None,
        dtype=None,
        dims=None,
        attrs=None,
        coords=None,
        indexes=None,
        name=None,
        mom_dims=None,
    ):
        """
        Parameters
        ----------
        dims : tuple, optional
            dimension names
        """
        kws, _, values = _check_xr_input(
            raw,
            axis=None,
            mom_dims=mom_dims,
            dims=dims,
            attrs=attrs,
            coords=coords,
            indexes=indexes,
            name=name,
        )

        return super(xCentralMoments, cls).from_raw(
            raw=values,
            mom_ndim=mom_ndim,
            mom=mom,
            val_shape=val_shape,
            dtype=dtype,
            **kws,
        )

    @classmethod
    def from_raws(
        cls,
        raws,
        mom=None,
        mom_ndim=None,
        axis=0,
        val_shape=None,
        dtype=None,
        dims=None,
        attrs=None,
        coords=None,
        indexes=None,
        name=None,
        mom_dims=None,
    ):

        kws, axis, values = _check_xr_input(
            raws,
            axis=axis,
            mom_dims=mom_dims,
            dims=dims,
            attrs=attrs,
            coords=coords,
            indexes=indexes,
            name=name,
        )

        super(xCentralMoments, cls).from_raws(
            values,
            mom=mom,
            mom_ndim=mom_ndim,
            axis=axis,
            val_shape=val_shape,
            dtype=dtype,
            **kws,
        )

    @classmethod
    def from_vals(
        cls,
        x,
        w=None,
        axis=0,
        mom=2,
        val_shape=None,
        dtype=None,
        broadcast=False,
        dims=None,
        attrs=None,
        coords=None,
        indexes=None,
        name=None,
        mom_dims=None,
    ):

        mom_ndim = cls._mom_ndim_from_mom(mom)
        x0 = x if mom_ndim == 1 else x[0]
        kws, axis, values = _check_xr_input(
            x0,
            axis=axis,
            mom_dims=mom_dims,
            dims=dims,
            attrs=attrs,
            coords=coords,
            indexes=indexes,
            name=name,
        )

        return super(xCentralMoments, cls).from_vals(
            x,
            w=w,
            axis=axis,
            mom=mom,
            val_shape=val_shape,
            dtype=dtype,
            broadcast=broadcast,
            **kws,
        )

    @classmethod
    def from_resample_vals(
        cls,
        x,
        freq=None,
        indices=None,
        nrep=None,
        w=None,
        axis=0,
        mom=2,
        rep_dim="rep",
        dtype=None,
        broadcast=False,
        parallel=True,
        resample_kws=None,
        dims=None,
        attrs=None,
        coords=None,
        indexes=None,
        name=None,
        mom_dims=None,
    ):

        mom_ndim = cls._mom_ndim_from_mom(mom)
        if mom_ndim == 1:
            y = None
        else:
            x, y = x

        kws, axis, values = _check_xr_input(
            x,
            axis=axis,
            mom_dims=mom_dims,
            dims=dims,
            attrs=attrs,
            coords=coords,
            indexes=indexes,
            name=name,
        )

        if kws["dims"] is not None:
            kws["dims"] = (rep_dim,) + tuple(kws["dims"])

        # reorder
        w = _xr_order_like(x, w)
        if y is not None:
            y = _xr_order_like(x, y)
            x = (x, y)

        return super(xCentralMoments, cls).from_resample_vals(
            x=x,
            freq=freq,
            indices=indices,
            nrep=nrep,
            w=w,
            axis=axis,
            mom=mom,
            dtype=dtype,
            broadcast=broadcast,
            parallel=parallel,
            resample_kws=resample_kws,
            **kws,
        )

    @classmethod
    def from_stat(
        cls,
        a,
        v=0.0,
        w=None,
        mom=2,
        val_shape=None,
        dtype=None,
        dims=None,
        attrs=None,
        coords=None,
        indexes=None,
        name=None,
        mom_dims=None,
    ):

        kws, _, values = _check_xr_input(
            a,
            axis=None,
            mom_dims=mom_dims,
            dims=dims,
            attrs=attrs,
            coords=coords,
            indexes=indexes,
            name=name,
        )

        return super(xCentralMoments, cls).from_stat(
            a=a, v=v, w=w, mom=mom, val_shape=val_shape, dtype=dtype, **kws
        )

    @classmethod
    def from_stats(
        cls,
        a,
        v=0.0,
        w=None,
        axis=0,
        mom=2,
        val_shape=None,
        dtype=None,
        dims=None,
        attrs=None,
        coords=None,
        indexes=None,
        name=None,
        mom_dims=None,
    ):

        kws, axis, values = _check_xr_input(
            a,
            axis=axis,
            mom_dims=mom_dims,
            dims=dims,
            attrs=attrs,
            coords=coords,
            indexes=indexes,
            name=name,
        )

        return super(xCentralMoments, cls).from_stat(
            a=a, v=v, w=w, axis=axis, mom=mom, val_shape=val_shape, dtype=dtype, **kws
        )

    def transpose(self, *dims, transpose_coords=None, copy=False, **kws):
        # make sure dims are last
        dims = list(dims)
        for k in self.mom_dims:
            if k in dims:
                dims.pop(dims.index(k))
        dims = tuple(dims) + self.mom_dims

        values = (
            self.values.transpose(*dims, transpose_coords=transpose_coords)
            # make sure mom are last
            # .transpose(...,*self.mom_dims)
        )
        return type(self).from_data(values, mom_ndim=self.mom_ndim, copy=copy, **kws)



from .central import CentralMoments, central_moments
from .resample import (
    bootstrap_confidence_interval,
    randsamp_freq,
    xbootstrap_confidence_interval,
)
from .xcentral import xcentral_moments, xCentralMoments

__all__ = (
    CentralMoments,
    central_moments,
    xCentralMoments,
    xcentral_moments,
    bootstrap_confidence_interval,
    randsamp_freq,
    xbootstrap_confidence_interval,
)

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions
